import asyncio
import logging
import math
from decimal import Decimal
from typing import Dict, List, Optional, Union

from hummingbot.connector.connector_base import ConnectorBase
from hummingbot.core.data_type.common import OrderType, PositionAction, PriceType, TradeType
from hummingbot.core.data_type.order_candidate import OrderCandidate, PerpetualOrderCandidate
from hummingbot.core.event.events import (
    BuyOrderCompletedEvent,
    BuyOrderCreatedEvent,
    MarketOrderFailureEvent,
    OrderCancelledEvent,
    OrderFilledEvent,
    SellOrderCompletedEvent,
    SellOrderCreatedEvent,
)
from hummingbot.logger import HummingbotLogger
from hummingbot.strategy.script_strategy_base import ScriptStrategyBase
from hummingbot.strategy_v2.executors.executor_base import ExecutorBase
from hummingbot.strategy_v2.executors.grid_executor.data_types import GridExecutorConfig, GridLevel, GridLevelStates
from hummingbot.strategy_v2.models.base import RunnableStatus
from hummingbot.strategy_v2.models.executors import CloseType, TrackedOrder
from hummingbot.strategy_v2.utils.distributions import Distributions


class GridExecutor(ExecutorBase):
    _logger = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self, strategy: ScriptStrategyBase, config: GridExecutorConfig,
                 update_interval: float = 1.0, max_retries: int = 10):
        """
        Initialize the PositionExecutor instance.

        :param strategy: The strategy to be used by the PositionExecutor.
        :param config: The configuration for the PositionExecutor, subclass of PositionExecutoConfig.
        :param update_interval: The interval at which the PositionExecutor should be updated, defaults to 1.0.
        :param max_retries: The maximum number of retries for the PositionExecutor, defaults to 5.
        """
        self.config: GridExecutorConfig = config
        if config.triple_barrier_config.time_limit_order_type != OrderType.MARKET or \
                config.triple_barrier_config.stop_loss_order_type != OrderType.MARKET:
            error = "Only market orders are supported for time_limit and stop_loss"
            self.logger().error(error)
            raise ValueError(error)
        super().__init__(strategy=strategy, config=config, connectors=[config.connector_name],
                         update_interval=update_interval)
        self.open_order_price_type = PriceType.BestBid if config.side == TradeType.BUY else PriceType.BestAsk
        self.close_order_price_type = PriceType.BestAsk if config.side == TradeType.BUY else PriceType.BestBid
        self.close_order_side = TradeType.BUY if config.side == TradeType.SELL else TradeType.SELL
        self.trading_rules = self.get_trading_rules(self.config.connector_name, self.config.trading_pair)
        # Grid levels
        self.grid_levels = self._generate_grid_levels()
        self.levels_by_state = {state: [] for state in GridLevelStates}
        self._close_order: Optional[TrackedOrder] = None
        self._filled_orders = []
        self._failed_orders = []
        self._canceled_orders = []

        self.step = Decimal("0")
        self.position_break_even_price = Decimal("0")
        self.position_size_base = Decimal("0")
        self.position_size_quote = Decimal("0")
        self.position_fees_quote = Decimal("0")
        self.position_pnl_quote = Decimal("0")
        self.position_pnl_pct = Decimal("0")
        self.open_liquidity_placed = Decimal("0")
        self.close_liquidity_placed = Decimal("0")
        self.realized_buy_size_quote = Decimal("0")
        self.realized_sell_size_quote = Decimal("0")
        self.realized_imbalance_quote = Decimal("0")
        self.realized_fees_quote = Decimal("0")
        self.realized_pnl_quote = Decimal("0")
        self.realized_pnl_pct = Decimal("0")
        self.max_open_creation_timestamp = 0
        self.max_close_creation_timestamp = 0
        self._open_fee_in_base = False

        self._trailing_stop_trigger_pct: Optional[Decimal] = None
        self._current_retries = 0
        self._max_retries = max_retries

    @property
    def is_perpetual(self) -> bool:
        """
        Check if the exchange connector is perpetual.

        :return: True if the exchange connector is perpetual, False otherwise.
        """
        return self.is_perpetual_connector(self.config.connector_name)

    async def validate_sufficient_balance(self):
        mid_price = self.get_price(self.config.connector_name, self.config.trading_pair, PriceType.MidPrice)
        total_amount_base = self.config.total_amount_quote / mid_price
        if self.is_perpetual:
            order_candidate = PerpetualOrderCandidate(
                trading_pair=self.config.trading_pair,
                is_maker=self.config.triple_barrier_config.open_order_type.is_limit_type(),
                order_type=self.config.triple_barrier_config.open_order_type,
                order_side=self.config.side,
                amount=total_amount_base,
                price=mid_price,
                leverage=Decimal(self.config.leverage),
            )
        else:
            order_candidate = OrderCandidate(
                trading_pair=self.config.trading_pair,
                is_maker=self.config.triple_barrier_config.open_order_type.is_limit_type(),
                order_type=self.config.triple_barrier_config.open_order_type,
                order_side=self.config.side,
                amount=total_amount_base,
                price=mid_price,
            )
        adjusted_order_candidates = self.adjust_order_candidates(self.config.connector_name, [order_candidate])
        if adjusted_order_candidates[0].amount == Decimal("0"):
            self.close_type = CloseType.INSUFFICIENT_BALANCE
            self.logger().error("Not enough budget to open position.")
            self.stop()

    def _generate_grid_levels(self):
        grid_levels = []
        price = self.get_price(self.config.connector_name, self.config.trading_pair, PriceType.MidPrice)
        # Get minimum notional and base amount increment from trading rules
        min_notional = max(
            self.config.min_order_amount_quote,
            self.trading_rules.min_notional_size
        )
        min_base_increment = self.trading_rules.min_base_amount_increment
        # Add safety margin to minimum notional to account for price movements and quantization
        min_notional_with_margin = min_notional * Decimal("1.05")  # 20% margin for safety
        # Calculate minimum base amount that satisfies both min_notional and quantization
        min_base_amount = max(
            min_notional_with_margin / price,  # Minimum from notional requirement
            min_base_increment * Decimal(str(math.ceil(float(min_notional) / float(min_base_increment * price))))
        )
        # Quantize the minimum base amount
        min_base_amount = Decimal(
            str(math.ceil(float(min_base_amount) / float(min_base_increment)))) * min_base_increment
        # Verify the quantized amount meets minimum notional
        min_quote_amount = min_base_amount * price
        # Calculate grid range and minimum step size
        grid_range = (self.config.end_price - self.config.start_price) / self.config.start_price
        min_step_size = max(
            self.config.min_spread_between_orders,
            self.trading_rules.min_price_increment / price
        )
        # Calculate maximum possible levels based on total amount
        max_possible_levels = int(self.config.total_amount_quote / min_quote_amount)
        if max_possible_levels == 0:
            # If we can't even create one level, create a single level with minimum amount
            n_levels = 1
            quote_amount_per_level = min_quote_amount
        else:
            # Calculate optimal number of levels
            max_levels_by_step = int(grid_range / min_step_size)
            n_levels = min(max_possible_levels, max_levels_by_step)
            # Calculate quote amount per level ensuring it meets minimum after quantization
            base_amount_per_level = max(
                min_base_amount,
                Decimal(str(math.floor(float(self.config.total_amount_quote / (price * n_levels)) /
                                       float(min_base_increment)))) * min_base_increment
            )
            quote_amount_per_level = base_amount_per_level * price
            # Adjust number of levels if total amount would be exceeded
            n_levels = min(n_levels, int(float(self.config.total_amount_quote) / float(quote_amount_per_level)))
        # Ensure we have at least one level
        n_levels = max(1, n_levels)
        # Generate price levels with even distribution
        if n_levels > 1:
            prices = Distributions.linear(n_levels, float(self.config.start_price), float(self.config.end_price))
            self.step = grid_range / (n_levels - 1)
        else:
            # For single level, use mid-point of range
            mid_price = (self.config.start_price + self.config.end_price) / 2
            prices = [mid_price]
            self.step = grid_range
        take_profit = max(self.step, self.config.triple_barrier_config.take_profit) if self.config.coerce_tp_to_step else self.config.triple_barrier_config.take_profit
        # Create grid levels
        for i, price in enumerate(prices):
            grid_levels.append(
                GridLevel(
                    id=f"L{i}",
                    price=price,
                    amount_quote=quote_amount_per_level,
                    take_profit=take_profit,
                    side=self.config.side,
                    open_order_type=self.config.triple_barrier_config.open_order_type,
                    take_profit_order_type=self.config.triple_barrier_config.take_profit_order_type,
                )
            )
        # Log grid creation details
        self.logger().info(
            f"Created {len(grid_levels)} grid levels with "
            f"amount per level: {quote_amount_per_level:.4f} {self.config.trading_pair.split('-')[1]} "
            f"(base amount: {(quote_amount_per_level / price):.8f} {self.config.trading_pair.split('-')[0]})"
        )
        return grid_levels

    @property
    def end_time(self) -> Optional[float]:
        """
        Calculate the end time of the position based on the time limit

        :return: The end time of the position.
        """
        if not self.config.triple_barrier_config.time_limit:
            return None
        return self.config.timestamp + self.config.triple_barrier_config.time_limit

    @property
    def is_expired(self) -> bool:
        """
        Check if the position is expired.

        :return: True if the position is expired, False otherwise.
        """
        return self.end_time and self.end_time <= self._strategy.current_timestamp

    @property
    def is_trading(self):
        """
        Check if the position is trading.

        :return: True if the position is trading, False otherwise.
        """
        return self.status == RunnableStatus.RUNNING and self.position_size_quote > Decimal("0")

    @property
    def is_active(self):
        """
        Returns whether the executor is open or trading.
        """
        return self._status in [RunnableStatus.RUNNING, RunnableStatus.NOT_STARTED, RunnableStatus.SHUTTING_DOWN]

    async def control_task(self):
        """
        This method is responsible for controlling the task based on the status of the executor.

        :return: None
        """
        self.update_grid_levels()
        self.update_metrics()
        if self.status == RunnableStatus.RUNNING:
            if self.control_triple_barrier():
                self.cancel_open_orders()
                self._status = RunnableStatus.SHUTTING_DOWN
                return
            open_orders_to_create = self.get_open_orders_to_create()
            close_orders_to_create = self.get_close_orders_to_create()
            open_order_ids_to_cancel = self.get_open_order_ids_to_cancel()
            close_order_ids_to_cancel = self.get_close_order_ids_to_cancel()
            for level in open_orders_to_create:
                self.adjust_and_place_open_order(level)
            for level in close_orders_to_create:
                self.adjust_and_place_close_order(level)
            for orders_id_to_cancel in open_order_ids_to_cancel + close_order_ids_to_cancel:
                # TODO: Implement batch order cancel
                self._strategy.cancel(
                    connector_name=self.config.connector_name,
                    trading_pair=self.config.trading_pair,
                    order_id=orders_id_to_cancel
                )
        elif self.status == RunnableStatus.SHUTTING_DOWN:
            await self.control_shutdown_process()
        self.evaluate_max_retries()

    def early_stop(self, keep_position: bool = False):
        """
        This method allows strategy to stop the executor early.

        :return: None
        """
        self.cancel_open_orders()
        self._status = RunnableStatus.SHUTTING_DOWN
        self.close_type = CloseType.POSITION_HOLD if keep_position else CloseType.EARLY_STOP

    def update_grid_levels(self):
        self.levels_by_state = {state: [] for state in GridLevelStates}
        for level in self.grid_levels:
            level.update_state()
            self.levels_by_state[level.state].append(level)
        completed = self.levels_by_state[GridLevelStates.COMPLETE]
        # Get completed orders and store them in the filled orders list
        for level in completed:
            if level.active_open_order.order.completely_filled_event.is_set() and level.active_close_order.order.completely_filled_event.is_set():
                open_order = level.active_open_order.order.to_json()
                close_order = level.active_close_order.order.to_json()
                self._filled_orders.append(open_order)
                self._filled_orders.append(close_order)
                self.levels_by_state[GridLevelStates.COMPLETE].remove(level)
                level.reset_level()
                self.levels_by_state[GridLevelStates.NOT_ACTIVE].append(level)

    async def control_shutdown_process(self):
        """
        Control the shutdown process of the executor, handling held positions separately
        """
        self.close_timestamp = self._strategy.current_timestamp
        open_orders_completed = self.open_liquidity_placed == Decimal("0")
        close_orders_completed = self.close_liquidity_placed == Decimal("0")

        if open_orders_completed and close_orders_completed:
            if self.close_type == CloseType.POSITION_HOLD:
                # Move filled orders to held positions instead of regular filled orders
                for level in self.levels_by_state[GridLevelStates.OPEN_ORDER_FILLED]:
                    if level.active_open_order and level.active_open_order.order:
                        self._held_position_orders.append(level.active_open_order.order.to_json())
                    level.reset_level()
                for level in self.levels_by_state[GridLevelStates.CLOSE_ORDER_PLACED]:
                    if level.active_close_order and level.active_close_order.order:
                        self._held_position_orders.append(level.active_close_order.order.to_json())
                    level.reset_level()
                if len(self._held_position_orders) == 0:
                    self.close_type = CloseType.EARLY_STOP
                self.levels_by_state = {}
                self.stop()
            else:
                # Regular shutdown process for non-held positions
                order_execution_completed = self.position_size_base == Decimal("0")
                if order_execution_completed:
                    for level in self.levels_by_state[GridLevelStates.OPEN_ORDER_FILLED]:
                        if level.active_open_order and level.active_open_order.order:
                            self._filled_orders.append(level.active_open_order.order.to_json())
                        level.reset_level()
                    for level in self.levels_by_state[GridLevelStates.CLOSE_ORDER_PLACED]:
                        if level.active_close_order and level.active_close_order.order:
                            self._filled_orders.append(level.active_close_order.order.to_json())
                        level.reset_level()
                    if self._close_order and self._close_order.order:
                        self._filled_orders.append(self._close_order.order.to_json())
                        self._close_order = None
                    self.update_realized_pnl_metrics()
                    self.levels_by_state = {}
                    self.stop()
                else:
                    await self.control_close_order()
                    self._current_retries += 1
        else:
            self.cancel_open_orders()
        await self._sleep(5.0)

    async def control_close_order(self):
        """
        This method is responsible for controlling the close order. If the close order is filled and the open orders are
        completed, it stops the executor. If the close order is not placed, it places the close order. If the close order
        is not filled, it waits for the close order to be filled and requests the order information to the connector.
        """
        if self._close_order:
            in_flight_order = self.get_in_flight_order(self.config.connector_name,
                                                       self._close_order.order_id) if not self._close_order.order else self._close_order.order
            if in_flight_order:
                self._close_order.order = in_flight_order
                self.logger().info("Waiting for close order to be filled")
            else:
                self._failed_orders.append(self._close_order.order_id)
                self._close_order = None
        elif not self.config.keep_position or self.close_type == CloseType.TAKE_PROFIT:
            self.place_close_order_and_cancel_open_orders(close_type=self.close_type)

    def adjust_and_place_open_order(self, level: GridLevel):
        """
        This method is responsible for adjusting the open order and placing it.

        :param level: The level to adjust and place the open order.
        :return: None
        """
        order_candidate = self._get_open_order_candidate(level)
        self.adjust_order_candidates(self.config.connector_name, [order_candidate])
        if order_candidate.amount > 0:
            order_id = self.place_order(
                connector_name=self.config.connector_name,
                trading_pair=self.config.trading_pair,
                order_type=self.config.triple_barrier_config.open_order_type,
                amount=order_candidate.amount,
                price=order_candidate.price,
                side=order_candidate.order_side,
                position_action=PositionAction.OPEN,
            )
            level.active_open_order = TrackedOrder(order_id=order_id)
            self.max_open_creation_timestamp = self._strategy.current_timestamp
            self.logger().debug(f"Executor ID: {self.config.id} - Placing open order {order_id}")

    def adjust_and_place_close_order(self, level: GridLevel):
        order_candidate = self._get_close_order_candidate(level)
        self.adjust_order_candidates(self.config.connector_name, [order_candidate])
        if order_candidate.amount > 0:
            order_id = self.place_order(
                connector_name=self.config.connector_name,
                trading_pair=self.config.trading_pair,
                order_type=self.config.triple_barrier_config.take_profit_order_type,
                amount=order_candidate.amount,
                price=order_candidate.price,
                side=order_candidate.order_side,
                position_action=PositionAction.CLOSE,
            )
            level.active_close_order = TrackedOrder(order_id=order_id)
            self.logger().debug(f"Executor ID: {self.config.id} - Placing close order {order_id}")

    def get_take_profit_price(self, level: GridLevel):
        return level.price * (1 + level.take_profit) if self.config.side == TradeType.BUY else level.price * (1 - level.take_profit)

    def _get_open_order_candidate(self, level: GridLevel):
        if ((level.side == TradeType.BUY and level.price >= self.current_open_quote) or
                (level.side == TradeType.SELL and level.price <= self.current_open_quote)):
            entry_price = self.current_open_quote * (1 - self.config.safe_extra_spread) if level.side == TradeType.BUY else self.current_open_quote * (1 + self.config.safe_extra_spread)
        else:
            entry_price = level.price
        if self.is_perpetual:
            return PerpetualOrderCandidate(
                trading_pair=self.config.trading_pair,
                is_maker=self.config.triple_barrier_config.open_order_type.is_limit_type(),
                order_type=self.config.triple_barrier_config.open_order_type,
                order_side=self.config.side,
                amount=level.amount_quote / self.mid_price,
                price=entry_price,
                leverage=Decimal(self.config.leverage)
            )
        return OrderCandidate(
            trading_pair=self.config.trading_pair,
            is_maker=self.config.triple_barrier_config.open_order_type.is_limit_type(),
            order_type=self.config.triple_barrier_config.open_order_type,
            order_side=self.config.side,
            amount=level.amount_quote / self.mid_price,
            price=entry_price
        )

    def _get_close_order_candidate(self, level: GridLevel):
        take_profit_price = self.get_take_profit_price(level)
        if ((level.side == TradeType.BUY and take_profit_price <= self.current_close_quote) or
                (level.side == TradeType.SELL and take_profit_price >= self.current_close_quote)):
            take_profit_price = self.current_close_quote * (
                1 + self.config.safe_extra_spread) if level.side == TradeType.BUY else self.current_close_quote * (
                1 - self.config.safe_extra_spread)
        if level.active_open_order.fee_asset == self.config.trading_pair.split("-")[0] and self.config.deduct_base_fees:
            amount = level.active_open_order.executed_amount_base - level.active_open_order.cum_fees_base
            self._open_fee_in_base = True
        else:
            amount = level.active_open_order.executed_amount_base
        if self.is_perpetual:
            return PerpetualOrderCandidate(
                trading_pair=self.config.trading_pair,
                is_maker=self.config.triple_barrier_config.take_profit_order_type.is_limit_type(),
                order_type=self.config.triple_barrier_config.take_profit_order_type,
                order_side=self.close_order_side,
                amount=amount,
                price=take_profit_price,
                leverage=Decimal(self.config.leverage)
            )
        return OrderCandidate(
            trading_pair=self.config.trading_pair,
            is_maker=self.config.triple_barrier_config.take_profit_order_type.is_limit_type(),
            order_type=self.config.triple_barrier_config.take_profit_order_type,
            order_side=self.close_order_side,
            amount=amount,
            price=take_profit_price
        )

    def update_metrics(self):
        self.mid_price = self.get_price(self.config.connector_name, self.config.trading_pair, PriceType.MidPrice)
        self.current_open_quote = self.get_price(self.config.connector_name, self.config.trading_pair,
                                                 price_type=self.open_order_price_type)
        self.current_close_quote = self.get_price(self.config.connector_name, self.config.trading_pair,
                                                  price_type=self.close_order_price_type)
        self.update_position_metrics()
        self.update_realized_pnl_metrics()

    def get_open_orders_to_create(self):
        """
        This method is responsible for controlling the open orders. Will check for each grid level if the order if there
        is an open order. If not, it will place a new orders from the proposed grid levels based on the current price,
        max open orders, max orders per batch, activation bounds and order frequency.
        """
        n_open_orders = len(
            [level.active_open_order for level in self.levels_by_state[GridLevelStates.OPEN_ORDER_PLACED]])
        if (self.max_open_creation_timestamp > self._strategy.current_timestamp - self.config.order_frequency or
                n_open_orders >= self.config.max_open_orders):
            return []
        levels_allowed = self._filter_levels_by_activation_bounds()
        sorted_levels_by_proximity = self._sort_levels_by_proximity(levels_allowed)
        return sorted_levels_by_proximity[:self.config.max_orders_per_batch]

    def get_close_orders_to_create(self):
        """
        This method is responsible for controlling the take profit. It will check if the net pnl percentage is greater
        than the take profit percentage and place the close order.

        :return: None
        """
        close_orders_proposal = []
        open_orders_filled = self.levels_by_state[GridLevelStates.OPEN_ORDER_FILLED]
        for level in open_orders_filled:
            if self.config.activation_bounds:
                tp_to_mid = abs(self.get_take_profit_price(level) - self.mid_price) / self.mid_price
                if tp_to_mid < self.config.activation_bounds:
                    close_orders_proposal.append(level)
            else:
                close_orders_proposal.append(level)
        return close_orders_proposal

    def get_open_order_ids_to_cancel(self):
        if self.config.activation_bounds:
            open_orders_to_cancel = []
            open_orders_placed = [level.active_open_order for level in
                                  self.levels_by_state[GridLevelStates.OPEN_ORDER_PLACED]]
            for order in open_orders_placed:
                price = order.price
                if price:
                    distance_pct = abs(price - self.mid_price) / self.mid_price
                    if distance_pct > self.config.activation_bounds:
                        open_orders_to_cancel.append(order.order_id)
                        self.logger().debug(f"Executor ID: {self.config.id} - Canceling open order {order.order_id}")
            return open_orders_to_cancel
        return []

    def get_close_order_ids_to_cancel(self):
        """
        This method is responsible for controlling the close orders. It will check if the take profit is greater than the
        current price and cancel the close order.

        :return: None
        """
        if self.config.activation_bounds:
            close_orders_to_cancel = []
            close_orders_placed = [level.active_close_order for level in
                                   self.levels_by_state[GridLevelStates.CLOSE_ORDER_PLACED]]
            for order in close_orders_placed:
                price = order.price
                if price:
                    distance_to_mid = abs(price - self.mid_price) / self.mid_price
                    if distance_to_mid > self.config.activation_bounds:
                        close_orders_to_cancel.append(order.order_id)
            return close_orders_to_cancel
        return []

    def _filter_levels_by_activation_bounds(self):
        not_active_levels = self.levels_by_state[GridLevelStates.NOT_ACTIVE]
        if self.config.activation_bounds:
            if self.config.side == TradeType.BUY:
                activation_bounds_price = self.mid_price * (1 - self.config.activation_bounds)
                return [level for level in not_active_levels if level.price >= activation_bounds_price]
            else:
                activation_bounds_price = self.mid_price * (1 + self.config.activation_bounds)
                return [level for level in not_active_levels if level.price <= activation_bounds_price]
        return not_active_levels

    def _sort_levels_by_proximity(self, levels: List[GridLevel]):
        return sorted(levels, key=lambda level: abs(level.price - self.mid_price))

    def control_triple_barrier(self):
        """
        This method is responsible for controlling the barriers. It controls the stop loss, take profit, time limit and
        trailing stop.

        :return: None
        """
        if self.stop_loss_condition():
            self.close_type = CloseType.STOP_LOSS
            return True
        elif self.limit_price_condition():
            self.close_type = CloseType.POSITION_HOLD if self.config.keep_position else CloseType.STOP_LOSS
            return True
        elif self.is_expired:
            self.close_type = CloseType.TIME_LIMIT
            return True
        elif self.trailing_stop_condition():
            self.close_type = CloseType.TRAILING_STOP
            return True
        elif self.take_profit_condition():
            self.close_type = CloseType.TAKE_PROFIT
            return True
        return False

    def take_profit_condition(self):
        """
        Take profit will be when the mid price is above the end price of the grid and there are no active executors.
        """
        if self.mid_price > self.config.end_price if self.config.side == TradeType.BUY else self.mid_price < self.config.start_price:
            return True
        return False

    def stop_loss_condition(self):
        """
        This method is responsible for controlling the stop loss. If the net pnl percentage is less than the stop loss
        percentage, it places the close order and cancels the open orders.

        :return: None
        """
        if self.config.triple_barrier_config.stop_loss:
            return self.position_pnl_pct <= -self.config.triple_barrier_config.stop_loss
        return False

    def limit_price_condition(self):
        """
        This method is responsible for controlling the limit price. If the current price is greater than the limit price,
        it places the close order and cancels the open orders.

        :return: None
        """
        if self.config.limit_price:
            if self.config.side == TradeType.BUY:
                return self.mid_price <= self.config.limit_price
            else:
                return self.mid_price >= self.config.limit_price
        return False

    def trailing_stop_condition(self):
        if self.config.triple_barrier_config.trailing_stop:
            net_pnl_pct = self.position_pnl_pct
            if not self._trailing_stop_trigger_pct:
                if net_pnl_pct > self.config.triple_barrier_config.trailing_stop.activation_price:
                    self._trailing_stop_trigger_pct = net_pnl_pct - self.config.triple_barrier_config.trailing_stop.trailing_delta
            else:
                if net_pnl_pct < self._trailing_stop_trigger_pct:
                    return True
                if net_pnl_pct - self.config.triple_barrier_config.trailing_stop.trailing_delta > self._trailing_stop_trigger_pct:
                    self._trailing_stop_trigger_pct = net_pnl_pct - self.config.triple_barrier_config.trailing_stop.trailing_delta
        return False

    def place_close_order_and_cancel_open_orders(self, close_type: CloseType, price: Decimal = Decimal("NaN")):
        """
        This method is responsible for placing the close order and canceling the open orders. If the difference between
        the open filled amount and the close filled amount is greater than the minimum order size, it places the close
        order. It also cancels the open orders.

        :param close_type: The type of the close order.
        :param price: The price to be used in the close order.
        :return: None
        """
        self.cancel_open_orders()
        if self.position_size_base >= self.trading_rules.min_order_size:
            order_id = self.place_order(
                connector_name=self.config.connector_name,
                trading_pair=self.config.trading_pair,
                order_type=OrderType.MARKET,
                amount=self.position_size_base,
                price=price,
                side=self.close_order_side,
                position_action=PositionAction.CLOSE,
            )
            self._close_order = TrackedOrder(order_id=order_id)
            self.logger().debug(f"Executor ID: {self.config.id} - Placing close order {order_id}")
        self.close_type = close_type
        self._status = RunnableStatus.SHUTTING_DOWN

    def cancel_open_orders(self):
        """
        This method is responsible for canceling the open orders.

        :return: None
        """
        open_order_placed = [level.active_open_order for level in
                             self.levels_by_state[GridLevelStates.OPEN_ORDER_PLACED]]
        close_order_placed = [level.active_close_order for level in
                              self.levels_by_state[GridLevelStates.CLOSE_ORDER_PLACED]]
        for order in open_order_placed + close_order_placed:
            # TODO: Implement cancel batch orders
            if order:
                self._strategy.cancel(
                    connector_name=self.config.connector_name,
                    trading_pair=self.config.trading_pair,
                    order_id=order.order_id
                )
                self.logger().debug("Removing open order")
                self.logger().debug(f"Executor ID: {self.config.id} - Canceling open order {order.order_id}")

    def get_custom_info(self) -> Dict:
        held_position_value = sum([
            Decimal(order["executed_amount_quote"])
            for order in self._held_position_orders
        ])

        return {
            "levels_by_state": {key.name: value for key, value in self.levels_by_state.items()},
            "filled_orders": self._filled_orders,
            "held_position_orders": self._held_position_orders,
            "held_position_value": held_position_value,
            "failed_orders": self._failed_orders,
            "canceled_orders": self._canceled_orders,
            "realized_buy_size_quote": self.realized_buy_size_quote,
            "realized_sell_size_quote": self.realized_sell_size_quote,
            "realized_imbalance_quote": self.realized_imbalance_quote,
            "realized_fees_quote": self.realized_fees_quote,
            "realized_pnl_quote": self.realized_pnl_quote,
            "realized_pnl_pct": self.realized_pnl_pct,
            "position_size_quote": self.position_size_quote,
            "position_fees_quote": self.position_fees_quote,
            "break_even_price": self.position_break_even_price,
            "position_pnl_quote": self.position_pnl_quote,
            "open_liquidity_placed": self.open_liquidity_placed,
            "close_liquidity_placed": self.close_liquidity_placed,
        }

    async def on_start(self):
        """
        This method is responsible for starting the executor and validating if the position is expired. The base method
        validates if there is enough balance to place the open order.

        :return: None
        """
        await super().on_start()
        self.update_metrics()
        if self.control_triple_barrier():
            self.logger().error(f"Grid is already expired by {self.close_type}.")

            self._status = RunnableStatus.SHUTTING_DOWN

    def evaluate_max_retries(self):
        """
        This method is responsible for evaluating the maximum number of retries to place an order and stop the executor
        if the maximum number of retries is reached.

        :return: None
        """
        if self._current_retries > self._max_retries:
            self.close_type = CloseType.FAILED
            self.stop()

    def update_tracked_orders_with_order_id(self, order_id: str):
        """
        This method is responsible for updating the tracked orders with the information from the InFlightOrder, using
        the order_id as a reference.

        :param order_id: The order_id to be used as a reference.
        :return: None
        """
        self.update_grid_levels()
        in_flight_order = self.get_in_flight_order(self.config.connector_name, order_id)
        if in_flight_order:
            for level in self.grid_levels:
                if level.active_open_order and level.active_open_order.order_id == order_id:
                    level.active_open_order.order = in_flight_order
                if level.active_close_order and level.active_close_order.order_id == order_id:
                    level.active_close_order.order = in_flight_order
            if self._close_order and self._close_order.order_id == order_id:
                self._close_order.order = in_flight_order

    def process_order_created_event(self, _, market, event: Union[BuyOrderCreatedEvent, SellOrderCreatedEvent]):
        """
        This method is responsible for processing the order created event. Here we will update the TrackedOrder with the
        order_id.
        """
        self.update_tracked_orders_with_order_id(event.order_id)

    def process_order_filled_event(self, _, market, event: OrderFilledEvent):
        """
        This method is responsible for processing the order filled event. Here we will update the value of
        _total_executed_amount_backup, that can be used if the InFlightOrder
        is not available.
        """
        self.update_tracked_orders_with_order_id(event.order_id)

    def process_order_completed_event(self, _, market, event: Union[BuyOrderCompletedEvent, SellOrderCompletedEvent]):
        """
        This method is responsible for processing the order completed event. Here we will check if the id is one of the
        tracked orders and update the state
        """
        self.update_tracked_orders_with_order_id(event.order_id)

    def process_order_canceled_event(self, _, market: ConnectorBase, event: OrderCancelledEvent):
        """
        This method is responsible for processing the order canceled event
        """
        self.update_grid_levels()
        levels_open_order_placed = [level for level in self.levels_by_state[GridLevelStates.OPEN_ORDER_PLACED]]
        levels_close_order_placed = [level for level in self.levels_by_state[GridLevelStates.CLOSE_ORDER_PLACED]]
        for level in levels_open_order_placed:
            if event.order_id == level.active_open_order.order_id:
                self._canceled_orders.append(level.active_open_order.order_id)
                self.max_open_creation_timestamp = 0
                level.reset_open_order()
        for level in levels_close_order_placed:
            if event.order_id == level.active_close_order.order_id:
                self._canceled_orders.append(level.active_close_order.order_id)
                self.max_close_creation_timestamp = 0
                level.reset_close_order()
        if self._close_order and event.order_id == self._close_order.order_id:
            self._canceled_orders.append(self._close_order.order_id)
            self._close_order = None

    def process_order_failed_event(self, _, market, event: MarketOrderFailureEvent):
        """
        This method is responsible for processing the order failed event. Here we will add the InFlightOrder to the
        failed orders list.
        """
        self.update_grid_levels()
        levels_open_order_placed = [level for level in self.levels_by_state[GridLevelStates.OPEN_ORDER_PLACED]]
        levels_close_order_placed = [level for level in self.levels_by_state[GridLevelStates.CLOSE_ORDER_PLACED]]
        for level in levels_open_order_placed:
            if event.order_id == level.active_open_order.order_id:
                self._failed_orders.append(level.active_open_order.order_id)
                self.max_open_creation_timestamp = 0
                level.reset_open_order()
        for level in levels_close_order_placed:
            if event.order_id == level.active_close_order.order_id:
                self._failed_orders.append(level.active_close_order.order_id)
                self.max_close_creation_timestamp = 0
                level.reset_close_order()
        if self._close_order and event.order_id == self._close_order.order_id:
            self._failed_orders.append(self._close_order.order_id)
            self._close_order = None

    def update_position_metrics(self):
        """
        Calculate the unrealized pnl in quote asset

        :return: The unrealized pnl in quote asset.
        """
        open_filled_levels = self.levels_by_state[GridLevelStates.OPEN_ORDER_FILLED] + self.levels_by_state[
            GridLevelStates.CLOSE_ORDER_PLACED]
        side_multiplier = 1 if self.config.side == TradeType.BUY else -1
        executed_amount_base = Decimal(sum([level.active_open_order.order.amount for level in open_filled_levels]))
        if executed_amount_base == Decimal("0"):
            self.position_size_base = Decimal("0")
            self.position_size_quote = Decimal("0")
            self.position_fees_quote = Decimal("0")
            self.position_pnl_quote = Decimal("0")
            self.position_pnl_pct = Decimal("0")
            self.close_liquidity_placed = Decimal("0")
        else:
            self.position_break_even_price = sum(
                [level.active_open_order.order.price * level.active_open_order.order.amount
                 for level in open_filled_levels]) / executed_amount_base
            if self._open_fee_in_base:
                executed_amount_base -= sum([level.active_open_order.cum_fees_base for level in open_filled_levels])
            close_order_size_base = self._close_order.executed_amount_base if self._close_order and self._close_order.is_done else Decimal(
                "0")
            self.position_size_base = executed_amount_base - close_order_size_base
            self.position_size_quote = self.position_size_base * self.position_break_even_price
            self.position_fees_quote = Decimal(sum([level.active_open_order.cum_fees_quote for level in open_filled_levels]))
            self.position_pnl_quote = side_multiplier * ((self.mid_price - self.position_break_even_price) / self.position_break_even_price) * self.position_size_quote - self.position_fees_quote
            self.position_pnl_pct = self.position_pnl_quote / self.position_size_quote if self.position_size_quote > 0 else Decimal(
                "0")
            self.close_liquidity_placed = sum([level.amount_quote for level in self.levels_by_state[GridLevelStates.CLOSE_ORDER_PLACED] if level.active_close_order and level.active_close_order.executed_amount_base == Decimal("0")])
        if len(self.levels_by_state[GridLevelStates.OPEN_ORDER_PLACED]) > 0:
            self.open_liquidity_placed = sum([level.amount_quote for level in self.levels_by_state[GridLevelStates.OPEN_ORDER_PLACED] if level.active_open_order and level.active_open_order.executed_amount_base == Decimal("0")])
        else:
            self.open_liquidity_placed = Decimal("0")

    def update_realized_pnl_metrics(self):
        """
        Calculate the realized pnl in quote asset, excluding held positions
        """
        if len(self._filled_orders) == 0:
            self._reset_metrics()
            return
        # Calculate metrics only for fully closed trades (not held positions)
        regular_filled_orders = [order for order in self._filled_orders
                                 if order not in self._held_position_orders]
        if len(regular_filled_orders) == 0:
            self._reset_metrics()
            return
        if self._open_fee_in_base:
            self.realized_buy_size_quote = sum([
                Decimal(order["executed_amount_quote"]) - Decimal(order["cumulative_fee_paid_quote"])
                for order in regular_filled_orders if order["trade_type"] == TradeType.BUY.name
            ])
        else:
            self.realized_buy_size_quote = sum([
                Decimal(order["executed_amount_quote"])
                for order in regular_filled_orders if order["trade_type"] == TradeType.BUY.name
            ])
        self.realized_sell_size_quote = sum([
            Decimal(order["executed_amount_quote"])
            for order in regular_filled_orders if order["trade_type"] == TradeType.SELL.name
        ])
        self.realized_imbalance_quote = self.realized_buy_size_quote - self.realized_sell_size_quote
        self.realized_fees_quote = sum([
            Decimal(order["cumulative_fee_paid_quote"])
            for order in regular_filled_orders
        ])
        self.realized_pnl_quote = (
            self.realized_sell_size_quote -
            self.realized_buy_size_quote -
            self.realized_fees_quote
        )
        self.realized_pnl_pct = (
            self.realized_pnl_quote / self.realized_buy_size_quote
            if self.realized_buy_size_quote > 0 else Decimal("0")
        )

    def _reset_metrics(self):
        """Helper method to reset all PnL metrics"""
        self.realized_buy_size_quote = Decimal("0")
        self.realized_sell_size_quote = Decimal("0")
        self.realized_imbalance_quote = Decimal("0")
        self.realized_fees_quote = Decimal("0")
        self.realized_pnl_quote = Decimal("0")
        self.realized_pnl_pct = Decimal("0")

    def get_net_pnl_quote(self) -> Decimal:
        """
        Calculate the net pnl in quote asset

        :return: The net pnl in quote asset.
        """
        return self.position_pnl_quote + self.realized_pnl_quote if self.close_type != CloseType.POSITION_HOLD else self.realized_pnl_quote

    def get_cum_fees_quote(self) -> Decimal:
        """
        Calculate the cumulative fees in quote asset

        :return: The cumulative fees in quote asset.
        """
        return self.position_fees_quote + self.realized_fees_quote if self.close_type != CloseType.POSITION_HOLD else self.realized_fees_quote

    @property
    def filled_amount_quote(self) -> Decimal:
        """
        Calculate the total amount in quote asset

        :return: The total amount in quote asset.
        """
        matched_volume = self.realized_buy_size_quote + self.realized_sell_size_quote
        return self.position_size_quote + matched_volume if self.close_type != CloseType.POSITION_HOLD else matched_volume

    def get_net_pnl_pct(self) -> Decimal:
        """
        Calculate the net pnl percentage

        :return: The net pnl percentage.
        """
        return self.get_net_pnl_quote() / self.filled_amount_quote if self.filled_amount_quote > 0 else Decimal("0")

    async def _sleep(self, delay: float):
        """
        This method is responsible for sleeping the executor for a specific time.

        :param delay: The time to sleep.
        :return: None
        """
        await asyncio.sleep(delay)



import hashlib
import hmac
import time
from collections import OrderedDict
from typing import Any, Dict, Optional
from urllib.parse import urlencode

import hummingbot.connector.exchange.hashkey.hashkey_constants as CONSTANTS
from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTRequest, WSRequest


class HashkeyAuth(AuthBase):

    def __init__(self, api_key: str, secret_key: str, time_provider: TimeSynchronizer):
        self.api_key = api_key
        self.secret_key = secret_key
        self.time_provider = time_provider

    @staticmethod
    def keysort(dictionary: Dict[str, str]) -> Dict[str, str]:
        return OrderedDict(sorted(dictionary.items(), key=lambda t: t[0]))

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        """
        Adds the server time and the signature to the request, required for authenticated interactions. It also adds
        the required parameter in the request header.
        :param request: the request to be configured for authenticated interaction
        """
        request.params = self.add_auth_to_params(params=request.params)
        headers = {
            "X-HK-APIKEY": self.api_key,
            "INPUT-SOURCE": CONSTANTS.HBOT_BROKER_ID,
        }
        if request.headers is not None:
            headers.update(request.headers)
        request.headers = headers
        return request

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        """
        This method is intended to configure a websocket request to be authenticated. Hashkey does not use this
        functionality
        """
        return request  # pass-through

    def add_auth_to_params(self,
                           params: Optional[Dict[str, Any]]):
        timestamp = int(self.time_provider.time() * 1e3)
        request_params = params or {}
        request_params["timestamp"] = timestamp
        request_params = self.keysort(request_params)
        signature = self._generate_signature(params=request_params)
        request_params["signature"] = signature
        return request_params

    def _generate_signature(self, params: Dict[str, Any]) -> str:
        encoded_params_str = urlencode(params)
        digest = hmac.new(self.secret_key.encode("utf8"), encoded_params_str.encode("utf8"), hashlib.sha256).hexdigest()
        return digest

    def generate_ws_authentication_message(self):
        """
        Generates the authentication message to start receiving messages from
        the 3 private ws channels
        """
        expires = int((self.time_provider.time() + 10) * 1e3)
        _val = f'GET/realtime{expires}'
        signature = hmac.new(self.secret_key.encode("utf8"),
                             _val.encode("utf8"), hashlib.sha256).hexdigest()
        auth_message = {
            "op": "auth",
            "args": [self.api_key, expires, signature]
        }
        return auth_message

    def _time(self):
        return time.time()



from decimal import Decimal
from enum import Enum
from typing import NamedTuple


class OrderType(Enum):
    MARKET = 1
    LIMIT = 2
    LIMIT_MAKER = 3

    def is_limit_type(self):
        return self in (OrderType.LIMIT, OrderType.LIMIT_MAKER)


class OpenOrder(NamedTuple):
    client_order_id: str
    trading_pair: str
    price: Decimal
    amount: Decimal
    executed_amount: Decimal
    status: str
    order_type: OrderType
    is_buy: bool
    time: int
    exchange_order_id: str


class PositionAction(Enum):
    OPEN = "OPEN"
    CLOSE = "CLOSE"
    NIL = "NIL"


# For Derivatives Exchanges
class PositionSide(Enum):
    LONG = "LONG"
    SHORT = "SHORT"
    BOTH = "BOTH"


# For Derivatives Exchanges
class PositionMode(Enum):
    HEDGE = "HEDGE"
    ONEWAY = "ONEWAY"


class PriceType(Enum):
    MidPrice = 1
    BestBid = 2
    BestAsk = 3
    LastTrade = 4
    LastOwnTrade = 5
    InventoryCost = 6
    Custom = 7


class TradeType(Enum):
    BUY = 1
    SELL = 2
    RANGE = 3


class LPType(Enum):
    ADD = 1
    REMOVE = 2
    COLLECT = 3



import asyncio
import base64
import logging
import time
from abc import ABC, abstractmethod
from decimal import Decimal
from enum import Enum
from typing import Any, Callable, Dict, List, Mapping, Optional, Tuple, Union

from bidict import bidict
from google.protobuf import any_pb2
from grpc import RpcError
from pyinjective import Transaction
from pyinjective.composer import Composer, injective_exchange_tx_pb
from pyinjective.core.market import DerivativeMarket, SpotMarket
from pyinjective.core.token import Token

from hummingbot.connector.derivative.position import Position
from hummingbot.connector.exchange.injective_v2 import injective_constants as CONSTANTS
from hummingbot.connector.exchange.injective_v2.injective_events import InjectiveEvent
from hummingbot.connector.exchange.injective_v2.injective_market import (
    InjectiveDerivativeMarket,
    InjectiveSpotMarket,
    InjectiveToken,
)
from hummingbot.connector.gateway.common_types import CancelOrderResult, PlaceOrderResult
from hummingbot.connector.gateway.gateway_in_flight_order import GatewayInFlightOrder, GatewayPerpetualInFlightOrder
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import to_0x_hex
from hummingbot.core.api_throttler.async_throttler_base import AsyncThrottlerBase
from hummingbot.core.data_type.common import OrderType, PositionAction, PositionSide, TradeType
from hummingbot.core.data_type.funding_info import FundingInfo, FundingInfoUpdate
from hummingbot.core.data_type.in_flight_order import OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType
from hummingbot.core.data_type.trade_fee import TokenAmount, TradeFeeBase, TradeFeeSchema
from hummingbot.core.event.event_listener import EventListener
from hummingbot.core.event.events import (
    AccountEvent,
    BalanceUpdateEvent,
    MarketEvent,
    OrderBookDataSourceEvent,
    PositionUpdateEvent,
)
from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.utils.async_utils import safe_gather
from hummingbot.logger import HummingbotLogger


class InjectiveDataSource(ABC):
    _logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(HummingbotLogger.logger_name_for_class(cls))
        return cls._logger

    @property
    @abstractmethod
    def publisher(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def query_executor(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def throttler(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def portfolio_account_injective_address(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def portfolio_account_subaccount_id(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def trading_account_injective_address(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def injective_chain_id(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def fee_denom(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def portfolio_account_subaccount_index(self) -> int:
        raise NotImplementedError

    @property
    @abstractmethod
    def network_name(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def last_received_message_timestamp(self):
        raise NotImplementedError

    @abstractmethod
    async def composer(self) -> Composer:
        raise NotImplementedError

    @abstractmethod
    async def timeout_height(self) -> int:
        raise NotImplementedError

    @abstractmethod
    async def spot_market_and_trading_pair_map(self):
        raise NotImplementedError

    @abstractmethod
    async def spot_market_info_for_id(self, market_id: str):
        raise NotImplementedError

    @abstractmethod
    async def derivative_market_and_trading_pair_map(self):
        raise NotImplementedError

    @abstractmethod
    async def derivative_market_info_for_id(self, market_id: str):
        raise NotImplementedError

    @abstractmethod
    async def trading_pair_for_market(self, market_id: str):
        raise NotImplementedError

    @abstractmethod
    async def market_id_for_spot_trading_pair(self, trading_pair: str) -> str:
        raise NotImplementedError

    @abstractmethod
    async def market_id_for_derivative_trading_pair(self, trading_pair: str) -> str:
        raise NotImplementedError

    @abstractmethod
    async def spot_markets(self):
        raise NotImplementedError

    @abstractmethod
    async def derivative_markets(self):
        raise NotImplementedError

    @abstractmethod
    async def token(self, denom: str) -> InjectiveToken:
        raise NotImplementedError

    @abstractmethod
    def events_listening_tasks(self) -> List[asyncio.Task]:
        raise NotImplementedError

    @abstractmethod
    def add_listening_task(self, task: asyncio.Task):
        raise NotImplementedError

    @abstractmethod
    def configure_throttler(self, throttler: AsyncThrottlerBase):
        raise NotImplementedError

    @abstractmethod
    async def trading_account_sequence(self) -> int:
        raise NotImplementedError

    @abstractmethod
    async def trading_account_number(self) -> int:
        raise NotImplementedError

    @abstractmethod
    async def initialize_trading_account(self):
        raise NotImplementedError

    @abstractmethod
    async def update_markets(self):
        raise NotImplementedError

    @abstractmethod
    def real_tokens_spot_trading_pair(self, unique_trading_pair: str) -> str:
        raise NotImplementedError

    @abstractmethod
    def real_tokens_perpetual_trading_pair(self, unique_trading_pair: str) -> str:
        raise NotImplementedError

    @abstractmethod
    async def order_updates_for_transaction(
            self,
            transaction_hash: str,
            spot_orders: Optional[List[GatewayInFlightOrder]] = None,
            perpetual_orders: Optional[List[GatewayPerpetualInFlightOrder]] = None,
    ) -> List[OrderUpdate]:
        raise NotImplementedError

    @abstractmethod
    def supported_order_types(self) -> List[OrderType]:
        raise NotImplementedError

    def is_started(self):
        return len(self.events_listening_tasks()) > 0

    async def check_network(self) -> NetworkStatus:
        try:
            await self.query_executor.ping()
            status = NetworkStatus.CONNECTED
        except asyncio.CancelledError:
            raise
        except Exception:
            status = NetworkStatus.NOT_CONNECTED
        return status

    async def start(self, market_ids: List[str]):
        if not self.is_started():
            await self.initialize_trading_account()
            if not self.is_started():
                spot_market_ids = []
                derivative_market_ids = []
                spot_markets = []
                derivative_markets = []
                for market_id in market_ids:
                    if market_id in await self.spot_market_and_trading_pair_map():
                        market = await self.spot_market_info_for_id(market_id=market_id)
                        spot_markets.append(market)
                        spot_market_ids.append(market_id)
                    else:
                        market = await self.derivative_market_info_for_id(market_id=market_id)
                        derivative_markets.append(market)
                        derivative_market_ids.append(market_id)

                self.add_listening_task(asyncio.create_task(self._listen_to_chain_transactions()))
                self.add_listening_task(asyncio.create_task(self._listen_to_chain_updates(
                    spot_markets=spot_markets,
                    derivative_markets=derivative_markets,
                    subaccount_ids=[self.portfolio_account_subaccount_id]
                )))

                await self._initialize_timeout_height()

    async def stop(self):
        for task in self.events_listening_tasks():
            task.cancel()

    def add_listener(self, event_tag: Enum, listener: EventListener):
        self.publisher.add_listener(event_tag=event_tag, listener=listener)

    def remove_listener(self, event_tag: Enum, listener: EventListener):
        self.publisher.remove_listener(event_tag=event_tag, listener=listener)

    async def spot_trading_rules(self) -> List[TradingRule]:
        markets = await self.spot_markets()
        trading_rules = self._create_trading_rules(markets=markets)

        return trading_rules

    async def derivative_trading_rules(self) -> List[TradingRule]:
        markets = await self.derivative_markets()
        trading_rules = self._create_trading_rules(markets=markets)

        return trading_rules

    async def spot_order_book_snapshot(self, market_id: str, trading_pair: str) -> OrderBookMessage:
        async with self.throttler.execute_task(limit_id=CONSTANTS.SPOT_ORDERBOOK_LIMIT_ID):
            snapshot_data = await self.query_executor.get_spot_orderbook(market_id=market_id)

        market = await self.spot_market_info_for_id(market_id=market_id)
        bids = [(market.price_from_chain_format(chain_price=Decimal(price)),
                 market.quantity_from_chain_format(chain_quantity=Decimal(quantity)))
                for price, quantity, _ in snapshot_data["buys"]]
        asks = [(market.price_from_chain_format(chain_price=Decimal(price)),
                 market.quantity_from_chain_format(chain_quantity=Decimal(quantity)))
                for price, quantity, _ in snapshot_data["sells"]]
        snapshot_msg = OrderBookMessage(
            message_type=OrderBookMessageType.SNAPSHOT,
            content={
                "trading_pair": trading_pair,
                "update_id": snapshot_data["sequence"],
                "bids": bids,
                "asks": asks,
            },
            timestamp=snapshot_data["timestamp"] * 1e-3,
        )
        return snapshot_msg

    async def perpetual_order_book_snapshot(self, market_id: str, trading_pair: str) -> OrderBookMessage:
        async with self.throttler.execute_task(limit_id=CONSTANTS.DERIVATIVE_ORDERBOOK_LIMIT_ID):
            snapshot_data = await self.query_executor.get_derivative_orderbook(market_id=market_id)

        market = await self.derivative_market_info_for_id(market_id=market_id)
        bids = [(market.price_from_chain_format(chain_price=Decimal(price)),
                 market.quantity_from_chain_format(chain_quantity=Decimal(quantity)))
                for price, quantity, _ in snapshot_data["buys"]]
        asks = [(market.price_from_chain_format(chain_price=Decimal(price)),
                 market.quantity_from_chain_format(chain_quantity=Decimal(quantity)))
                for price, quantity, _ in snapshot_data["sells"]]
        snapshot_msg = OrderBookMessage(
            message_type=OrderBookMessageType.SNAPSHOT,
            content={
                "trading_pair": trading_pair,
                "update_id": snapshot_data["sequence"],
                "bids": bids,
                "asks": asks,
            },
            timestamp=snapshot_data["timestamp"] * 1e-3,
        )
        return snapshot_msg

    async def last_traded_price(self, market_id: str) -> Decimal:
        price = await self._last_traded_price(market_id=market_id)
        return price

    async def all_account_balances(self) -> Dict[str, Dict[str, Decimal]]:
        account_address = self.portfolio_account_injective_address

        async with self.throttler.execute_task(limit_id=CONSTANTS.PORTFOLIO_BALANCES_LIMIT_ID):
            portfolio_response = await self.query_executor.account_portfolio(account_address=account_address)

        bank_balances = portfolio_response["portfolio"]["bankBalances"]
        sub_account_balances = portfolio_response["portfolio"].get("subaccounts", [])

        balances_dict: Dict[str, Dict[str, Decimal]] = {}

        if self._uses_default_portfolio_subaccount():
            for bank_entry in bank_balances:
                token = await self.token(denom=bank_entry["denom"])
                if token is not None:
                    asset_name: str = token.unique_symbol

                    available_balance = token.value_from_chain_format(chain_value=Decimal(bank_entry["amount"]))
                    total_balance = available_balance
                    balances_dict[asset_name] = {
                        "total_balance": total_balance,
                        "available_balance": available_balance,
                    }

        for entry in sub_account_balances:
            if entry["subaccountId"] == self.portfolio_account_subaccount_id:
                token = await self.token(denom=entry["denom"])
                if token is not None:
                    asset_name: str = token.unique_symbol

                    total_balance = token.value_from_chain_format(chain_value=Decimal(entry["deposit"]["totalBalance"]))
                    available_balance = token.value_from_chain_format(
                        chain_value=Decimal(entry["deposit"]["availableBalance"]))

                    balance_element = balances_dict.get(
                        asset_name, {"total_balance": Decimal("0"), "available_balance": Decimal("0")}
                    )
                    balance_element["total_balance"] += total_balance
                    balance_element["available_balance"] += available_balance
                    balances_dict[asset_name] = balance_element

        return balances_dict

    async def account_positions(self) -> List[Position]:
        done = False
        skip = 0
        position_entries = []

        while not done:
            async with self.throttler.execute_task(limit_id=CONSTANTS.POSITIONS_LIMIT_ID):
                positions_response = await self.query_executor.get_derivative_positions(
                    subaccount_id=self.portfolio_account_subaccount_id,
                    skip=skip,
                )
            if "positions" in positions_response:
                total = int(positions_response["paging"]["total"])
                entries = positions_response["positions"]

                position_entries.extend(entries)
                done = len(position_entries) >= total
                skip += len(entries)
            else:
                done = True

        positions = []
        for position_entry in position_entries:
            position_update = await self._parse_position_update_event(event=position_entry)

            position = Position(
                trading_pair=position_update.trading_pair,
                position_side=position_update.position_side,
                unrealized_pnl=position_update.unrealized_pnl,
                entry_price=position_update.entry_price,
                amount=position_update.amount,
                leverage=position_update.leverage,
            )

            positions.append(position)

        return positions

    async def create_orders(
            self,
            spot_orders: Optional[List[GatewayInFlightOrder]] = None,
            perpetual_orders: Optional[List[GatewayPerpetualInFlightOrder]] = None,
    ) -> List[PlaceOrderResult]:
        spot_orders = spot_orders or []
        perpetual_orders = perpetual_orders or []
        results = []
        if len(spot_orders) > 0 or len(perpetual_orders) > 0:
            order_creation_messages = await self._order_creation_messages(
                spot_orders_to_create=spot_orders,
                derivative_orders_to_create=perpetual_orders,
            )

            try:
                result = await self._send_in_transaction(messages=order_creation_messages)
                if result["code"] != 0 or result["txhash"] in [None, ""]:
                    raise ValueError(f"Error sending the order creation transaction ({result['rawLog']})")
                else:
                    transaction_hash = result["txhash"]
                    results = self._place_order_results(
                        orders_to_create=spot_orders + perpetual_orders,
                        misc_updates={
                            "creation_transaction_hash": transaction_hash,
                        },
                    )
            except asyncio.CancelledError:
                raise
            except Exception as ex:
                self.logger().debug(
                    f"Error broadcasting transaction to create orders (message: {order_creation_messages})")
                results = self._place_order_results(
                    orders_to_create=spot_orders + perpetual_orders,
                    misc_updates={},
                    exception=ex,
                )

        return results

    async def cancel_orders(
            self,
            spot_orders: Optional[List[GatewayInFlightOrder]] = None,
            perpetual_orders: Optional[List[GatewayPerpetualInFlightOrder]] = None,
    ) -> List[CancelOrderResult]:
        spot_orders = spot_orders or []
        perpetual_orders = perpetual_orders or []

        orders_with_hash = []
        spot_orders_data = []
        derivative_orders_data = []
        results = []

        if len(spot_orders) > 0 or len(perpetual_orders) > 0:
            for order in spot_orders:
                market_id = await self.market_id_for_spot_trading_pair(trading_pair=order.trading_pair)
                order_data = await self._generate_injective_order_data(order=order, market_id=market_id)
                spot_orders_data.append(order_data)
                orders_with_hash.append(order)

            for order in perpetual_orders:
                market_id = await self.market_id_for_derivative_trading_pair(trading_pair=order.trading_pair)
                order_data = await self._generate_injective_order_data(order=order, market_id=market_id)
                derivative_orders_data.append(order_data)
                orders_with_hash.append(order)

            if len(orders_with_hash) > 0:
                delegated_message = await self._order_cancel_message(
                    spot_orders_to_cancel=spot_orders_data,
                    derivative_orders_to_cancel=derivative_orders_data,
                )

                try:
                    result = await self._send_in_transaction(messages=[delegated_message])
                    if result["code"] != 0:
                        raise ValueError(f"Error sending the order cancel transaction ({result['rawLog']})")
                    else:
                        cancel_transaction_hash = result.get("txhash", "")
                        results.extend([
                            CancelOrderResult(
                                client_order_id=order.client_order_id,
                                trading_pair=order.trading_pair,
                                misc_updates={"cancelation_transaction_hash": cancel_transaction_hash},
                            ) for order in orders_with_hash
                        ])
                except asyncio.CancelledError:
                    raise
                except Exception as ex:
                    self.logger().debug(f"Error broadcasting transaction to cancel orders (message: {delegated_message})")
                    results.extend([
                        CancelOrderResult(
                            client_order_id=order.client_order_id,
                            trading_pair=order.trading_pair,
                            exception=ex,
                        ) for order in orders_with_hash
                    ])

        return results

    async def cancel_all_subaccount_orders(
            self,
            spot_markets_ids: Optional[List[str]] = None,
            perpetual_markets_ids: Optional[List[str]] = None,
    ):
        spot_markets_ids = spot_markets_ids or []
        perpetual_markets_ids = perpetual_markets_ids or []

        delegated_message = await self._all_subaccount_orders_cancel_message(
            spot_markets_ids=spot_markets_ids,
            derivative_markets_ids=perpetual_markets_ids,
        )

        result = await self._send_in_transaction(messages=[delegated_message])
        if result["code"] != 0:
            raise ValueError(f"Error sending the order cancel transaction ({result['rawLog']})")

    async def spot_trade_updates(self, market_ids: List[str], start_time: float) -> List[TradeUpdate]:
        done = False
        skip = 0
        trade_entries = []

        while not done:
            async with self.throttler.execute_task(limit_id=CONSTANTS.SPOT_TRADES_LIMIT_ID):
                trades_response = await self.query_executor.get_spot_trades(
                    market_ids=market_ids,
                    subaccount_id=self.portfolio_account_subaccount_id,
                    start_time=int(start_time * 1e3),
                    skip=skip,
                )
            if "trades" in trades_response:
                total = int(trades_response["paging"]["total"])
                entries = trades_response["trades"]

                trade_entries.extend(entries)
                done = len(trade_entries) >= total
                skip += len(entries)
            else:
                done = True

        trade_updates = [await self._parse_spot_trade_entry(trade_info=trade_info) for trade_info in trade_entries]

        return trade_updates

    async def perpetual_trade_updates(self, market_ids: List[str], start_time: float) -> List[TradeUpdate]:
        done = False
        skip = 0
        trade_entries = []

        while not done:
            async with self.throttler.execute_task(limit_id=CONSTANTS.DERIVATIVE_TRADES_LIMIT_ID):
                trades_response = await self.query_executor.get_derivative_trades(
                    market_ids=market_ids,
                    subaccount_id=self.portfolio_account_subaccount_id,
                    start_time=int(start_time * 1e3),
                    skip=skip,
                )
            if "trades" in trades_response:
                total = int(trades_response["paging"]["total"])
                entries = trades_response["trades"]

                trade_entries.extend(entries)
                done = len(trade_entries) >= total
                skip += len(entries)
            else:
                done = True

        trade_updates = [await self._parse_derivative_trade_entry(trade_info=trade_info) for trade_info in trade_entries]

        return trade_updates

    async def spot_order_updates(self, market_ids: List[str], start_time: float) -> List[OrderUpdate]:
        done = False
        skip = 0
        order_entries = []

        while not done:
            async with self.throttler.execute_task(limit_id=CONSTANTS.SPOT_ORDERS_HISTORY_LIMIT_ID):
                orders_response = await self.query_executor.get_historical_spot_orders(
                    market_ids=market_ids,
                    subaccount_id=self.portfolio_account_subaccount_id,
                    start_time=int(start_time * 1e3),
                    skip=skip,
                )
            if "orders" in orders_response:
                total = int(orders_response["paging"]["total"])
                entries = orders_response["orders"]

                order_entries.extend(entries)
                done = len(order_entries) >= total
                skip += len(entries)
            else:
                done = True

        order_updates = [await self._parse_order_entry(order_info=order_info) for order_info in order_entries]

        return order_updates

    async def perpetual_order_updates(self, market_ids: List[str], start_time: float) -> List[OrderUpdate]:
        done = False
        skip = 0
        order_entries = []

        while not done:
            async with self.throttler.execute_task(limit_id=CONSTANTS.DERIVATIVE_ORDERS_HISTORY_LIMIT_ID):
                orders_response = await self.query_executor.get_historical_derivative_orders(
                    market_ids=market_ids,
                    subaccount_id=self.portfolio_account_subaccount_id,
                    start_time=int(start_time * 1e3),
                    skip=skip,
                )
            if "orders" in orders_response:
                total = int(orders_response["paging"]["total"])
                entries = orders_response["orders"]

                order_entries.extend(entries)
                done = len(order_entries) >= total
                skip += len(entries)
            else:
                done = True

        order_updates = [await self._parse_order_entry(order_info=order_info) for order_info in order_entries]

        return order_updates

    async def get_spot_trading_fees(self) -> Dict[str, TradeFeeSchema]:
        markets = await self.spot_markets()
        fees = await self._create_trading_fees(markets=markets)

        return fees

    async def get_derivative_trading_fees(self) -> Dict[str, TradeFeeSchema]:
        markets = await self.derivative_markets()
        fees = await self._create_trading_fees(markets=markets)

        return fees

    async def funding_info(self, market_id: str) -> FundingInfo:
        funding_rate = await self.last_funding_rate(market_id=market_id)
        oracle_price = await self._oracle_price(market_id=market_id)
        last_traded_price = await self.last_traded_price(market_id=market_id)
        updated_market_info = await self._updated_derivative_market_info_for_id(market_id=market_id)

        funding_info = FundingInfo(
            trading_pair=await self.trading_pair_for_market(market_id=market_id),
            index_price=last_traded_price,  # Use the last traded price as the index_price
            mark_price=oracle_price,
            next_funding_utc_timestamp=int(updated_market_info["market"]["perpetualMarketInfo"]["nextFundingTimestamp"]),
            rate=funding_rate,
        )
        return funding_info

    async def last_funding_rate(self, market_id: str) -> Decimal:
        async with self.throttler.execute_task(limit_id=CONSTANTS.FUNDING_RATES_LIMIT_ID):
            response = await self.query_executor.get_funding_rates(market_id=market_id, limit=1)
        funding_rates = response.get("fundingRates", [])
        if len(funding_rates) == 0:
            rate = Decimal("0")
        else:
            rate = Decimal(response["fundingRates"][0]["rate"])

        return rate

    async def last_funding_payment(self, market_id: str) -> Tuple[Decimal, float]:
        async with self.throttler.execute_task(limit_id=CONSTANTS.FUNDING_PAYMENTS_LIMIT_ID):
            response = await self.query_executor.get_funding_payments(
                subaccount_id=self.portfolio_account_subaccount_id,
                market_id=market_id,
                limit=1
            )

        last_payment = Decimal(-1)
        last_timestamp = 0
        payments = response.get("payments", [])

        if len(payments) > 0:
            last_payment = Decimal(payments[0]["amount"])
            last_timestamp = int(payments[0]["timestamp"]) * 1e-3

        return last_payment, last_timestamp

    @abstractmethod
    async def _initialize_timeout_height(self):
        raise NotImplementedError

    @abstractmethod
    def _sign_and_encode(self, transaction: Transaction) -> bytes:
        raise NotImplementedError

    @abstractmethod
    def _uses_default_portfolio_subaccount(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    async def _order_creation_messages(
            self,
            spot_orders_to_create: List[GatewayInFlightOrder],
            derivative_orders_to_create: List[GatewayPerpetualInFlightOrder],
    ) -> List[any_pb2.Any]:
        raise NotImplementedError

    @abstractmethod
    async def _order_cancel_message(
            self,
            spot_orders_to_cancel: List[injective_exchange_tx_pb.OrderData],
            derivative_orders_to_cancel: List[injective_exchange_tx_pb.OrderData]
    ) -> any_pb2.Any:
        raise NotImplementedError

    @abstractmethod
    async def _all_subaccount_orders_cancel_message(
            self,
            spot_markets_ids: List[str],
            derivative_markets_ids: List[str]
    ) -> any_pb2.Any:
        raise NotImplementedError

    @abstractmethod
    async def _generate_injective_order_data(self, order: GatewayInFlightOrder, market_id: str) -> injective_exchange_tx_pb.OrderData:
        raise NotImplementedError

    @abstractmethod
    async def _updated_derivative_market_info_for_id(self, market_id: str) -> Dict[str, Any]:
        raise NotImplementedError

    @abstractmethod
    async def _configure_gas_fee_for_transaction(self, transaction: Transaction):
        raise NotImplementedError

    def _place_order_results(
            self,
            orders_to_create: List[GatewayInFlightOrder],
            misc_updates: Dict[str, Any],
            exception: Optional[Exception] = None,
    ) -> List[PlaceOrderResult]:
        return [
            PlaceOrderResult(
                update_timestamp=self._time(),
                client_order_id=order.client_order_id,
                exchange_order_id=None,
                trading_pair=order.trading_pair,
                misc_updates=misc_updates,
                exception=exception
            ) for order in orders_to_create
        ]

    async def _last_traded_price(self, market_id: str) -> Decimal:
        price = Decimal("nan")
        if market_id in await self.spot_market_and_trading_pair_map():
            market = await self.spot_market_info_for_id(market_id=market_id)
            async with self.throttler.execute_task(limit_id=CONSTANTS.SPOT_TRADES_LIMIT_ID):
                trades_response = await self.query_executor.get_spot_trades(
                    market_ids=[market_id],
                    limit=1,
                )
            trades = trades_response.get("trades", [])
            if len(trades) > 0:
                price = market.price_from_chain_format(
                    chain_price=Decimal(trades[0]["price"]["price"]))

        else:
            market = await self.derivative_market_info_for_id(market_id=market_id)
            async with self.throttler.execute_task(limit_id=CONSTANTS.DERIVATIVE_TRADES_LIMIT_ID):
                trades_response = await self.query_executor.get_derivative_trades(
                    market_ids=[market_id],
                    limit=1,
                )
            trades = trades_response.get("trades", [])
            if len(trades) > 0:
                price = market.price_from_chain_format(
                    chain_price=Decimal(trades_response["trades"][0]["positionDelta"]["executionPrice"]))

        return price

    async def _oracle_price(self, market_id: str) -> Decimal:
        market = await self.derivative_market_info_for_id(market_id=market_id)
        async with self.throttler.execute_task(limit_id=CONSTANTS.ORACLE_PRICES_LIMIT_ID):
            response = await self.query_executor.get_oracle_prices(
                base_symbol=market.oracle_base(),
                quote_symbol=market.oracle_quote(),
                oracle_type=market.oracle_type(),
                oracle_scale_factor=0,
            )
        price = Decimal(response["price"])

        return price

    async def _listen_chain_stream_updates(
            self,
            spot_markets: List[InjectiveSpotMarket],
            derivative_markets: List[InjectiveDerivativeMarket],
            subaccount_ids: List[str],
            composer: Composer,
            callback: Callable,
            on_end_callback: Optional[Callable] = None,
            on_status_callback: Optional[Callable] = None,
    ):
        spot_market_ids = [market_info.market_id for market_info in spot_markets]
        derivative_market_ids = []
        oracle_price_symbols = set()

        for derivative_market_info in derivative_markets:
            derivative_market_ids.append(derivative_market_info.market_id)
            oracle_price_symbols.add(derivative_market_info.oracle_base())
            oracle_price_symbols.add(derivative_market_info.oracle_quote())

        subaccount_deposits_filter = composer.chain_stream_subaccount_deposits_filter(subaccount_ids=subaccount_ids)
        if len(spot_market_ids) > 0:
            spot_orderbooks_filter = composer.chain_stream_orderbooks_filter(market_ids=spot_market_ids)
            spot_trades_filter = composer.chain_stream_trades_filter(market_ids=spot_market_ids)
            spot_orders_filter = composer.chain_stream_orders_filter(
                subaccount_ids=subaccount_ids, market_ids=spot_market_ids,
            )
        else:
            spot_orderbooks_filter = None
            spot_trades_filter = None
            spot_orders_filter = None

        if len(derivative_market_ids) > 0:
            derivative_orderbooks_filter = composer.chain_stream_orderbooks_filter(market_ids=derivative_market_ids)
            derivative_trades_filter = composer.chain_stream_trades_filter(market_ids=derivative_market_ids)
            derivative_orders_filter = composer.chain_stream_orders_filter(
                subaccount_ids=subaccount_ids, market_ids=derivative_market_ids
            )
            positions_filter = composer.chain_stream_positions_filter(
                subaccount_ids=subaccount_ids, market_ids=derivative_market_ids
            )
            oracle_price_filter = composer.chain_stream_oracle_price_filter(symbols=list(oracle_price_symbols))
        else:
            derivative_orderbooks_filter = None
            derivative_trades_filter = None
            derivative_orders_filter = None
            positions_filter = None
            oracle_price_filter = None

        await self.query_executor.listen_chain_stream_updates(
            callback=callback,
            on_end_callback=on_end_callback,
            on_status_callback=on_status_callback,
            subaccount_deposits_filter=subaccount_deposits_filter,
            spot_trades_filter=spot_trades_filter,
            derivative_trades_filter=derivative_trades_filter,
            spot_orders_filter=spot_orders_filter,
            derivative_orders_filter=derivative_orders_filter,
            spot_orderbooks_filter=spot_orderbooks_filter,
            derivative_orderbooks_filter=derivative_orderbooks_filter,
            positions_filter=positions_filter,
            oracle_price_filter=oracle_price_filter
        )

    async def _listen_transactions_updates(
        self,
        callback: Callable,
        on_end_callback: Callable,
        on_status_callback: Callable,
    ):
        await self.query_executor.listen_transactions_updates(
            callback=callback,
            on_end_callback=on_end_callback,
            on_status_callback=on_status_callback,
        )

    async def _parse_spot_trade_entry(self, trade_info: Dict[str, Any]) -> TradeUpdate:
        exchange_order_id: str = trade_info["orderHash"]
        client_order_id: str = trade_info.get("cid", "")
        market = await self.spot_market_info_for_id(market_id=trade_info["marketId"])
        trading_pair = await self.trading_pair_for_market(market_id=trade_info["marketId"])

        price = market.price_from_chain_format(chain_price=Decimal(trade_info["price"]["price"]))
        size = market.quantity_from_chain_format(chain_quantity=Decimal(trade_info["price"]["quantity"]))
        trade_type = TradeType.BUY if trade_info["tradeDirection"] == "buy" else TradeType.SELL
        is_taker: bool = trade_info["executionSide"] == "taker"
        trade_time = int(trade_info["executedAt"]) * 1e-3
        trade_id = trade_info["tradeId"]

        fee_amount = market.quote_token.value_from_chain_format(chain_value=Decimal(trade_info["fee"]))
        fee = TradeFeeBase.new_spot_fee(
            fee_schema=TradeFeeSchema(),
            trade_type=trade_type,
            percent_token=market.quote_token.symbol,
            flat_fees=[TokenAmount(amount=fee_amount, token=market.quote_token.symbol)]
        )

        trade_update = TradeUpdate(
            trade_id=trade_id,
            client_order_id=client_order_id,
            exchange_order_id=exchange_order_id,
            trading_pair=trading_pair,
            fill_timestamp=trade_time,
            fill_price=price,
            fill_base_amount=size,
            fill_quote_amount=size * price,
            fee=fee,
            is_taker=is_taker,
        )

        return trade_update

    async def _parse_derivative_trade_entry(self, trade_info: Dict[str, Any]) -> TradeUpdate:
        exchange_order_id: str = trade_info["orderHash"]
        client_order_id: str = trade_info.get("cid", "")
        market = await self.derivative_market_info_for_id(market_id=trade_info["marketId"])
        trading_pair = await self.trading_pair_for_market(market_id=trade_info["marketId"])

        price = market.price_from_chain_format(chain_price=Decimal(trade_info["positionDelta"]["executionPrice"]))
        size = market.quantity_from_chain_format(chain_quantity=Decimal(trade_info["positionDelta"]["executionQuantity"]))
        is_taker: bool = trade_info["executionSide"] == "taker"
        trade_time = int(trade_info["executedAt"]) * 1e-3
        trade_id = trade_info["tradeId"]

        fee_amount = market.quote_token.value_from_chain_format(chain_value=Decimal(trade_info["fee"]))
        fee = TradeFeeBase.new_perpetual_fee(
            fee_schema=TradeFeeSchema(),
            position_action=PositionAction.OPEN,  # will be changed by the exchange class
            percent_token=market.quote_token.symbol,
            flat_fees=[TokenAmount(amount=fee_amount, token=market.quote_token.symbol)]
        )

        trade_update = TradeUpdate(
            trade_id=trade_id,
            client_order_id=client_order_id,
            exchange_order_id=exchange_order_id,
            trading_pair=trading_pair,
            fill_timestamp=trade_time,
            fill_price=price,
            fill_base_amount=size,
            fill_quote_amount=size * price,
            fee=fee,
            is_taker=is_taker,
        )

        return trade_update

    async def _parse_order_entry(self, order_info: Dict[str, Any]) -> OrderUpdate:
        exchange_order_id: str = order_info["orderHash"]
        client_order_id: str = order_info.get("cid", "")
        trading_pair = await self.trading_pair_for_market(market_id=order_info["marketId"])

        status_update = OrderUpdate(
            trading_pair=trading_pair,
            update_timestamp=int(order_info["updatedAt"]) * 1e-3,
            new_state=CONSTANTS.ORDER_STATE_MAP[order_info["state"]],
            client_order_id=client_order_id,
            exchange_order_id=exchange_order_id,
        )

        return status_update

    async def _parse_position_update_event(self, event: Dict[str, Any]) -> PositionUpdateEvent:
        market = await self.derivative_market_info_for_id(market_id=event["marketId"])
        trading_pair = await self.trading_pair_for_market(market_id=event["marketId"])

        if "direction" in event:
            position_side = PositionSide[event["direction"].upper()]
            amount_sign = Decimal(-1) if position_side == PositionSide.SHORT else Decimal(1)
            chain_entry_price = Decimal(event["entryPrice"])
            chain_mark_price = Decimal(event["markPrice"])
            chain_amount = Decimal(event["quantity"])
            chain_margin = Decimal(event["margin"])
            entry_price = market.price_from_chain_format(chain_price=chain_entry_price)
            mark_price = market.price_from_chain_format(chain_price=chain_mark_price)
            amount = market.quantity_from_chain_format(chain_quantity=chain_amount)
            leverage = (chain_amount * chain_entry_price) / chain_margin
            unrealized_pnl = (mark_price - entry_price) * amount * amount_sign
        else:
            position_side = None
            entry_price = unrealized_pnl = amount = Decimal("0")
            leverage = amount_sign = Decimal("1")

        parsed_event = PositionUpdateEvent(
            timestamp=int(event["updatedAt"]) * 1e-3,
            trading_pair=trading_pair,
            position_side=position_side,
            unrealized_pnl=unrealized_pnl,
            entry_price=entry_price,
            amount=amount * amount_sign,
            leverage=leverage,
        )

        return parsed_event

    async def _send_in_transaction(self, messages: List[any_pb2.Any]) -> Dict[str, Any]:
        transaction = Transaction()
        transaction.with_messages(*messages)
        transaction.with_sequence(await self.trading_account_sequence())
        transaction.with_account_num(await self.trading_account_number())
        transaction.with_chain_id(self.injective_chain_id)

        async with self.throttler.execute_task(limit_id=CONSTANTS.SIMULATE_TRANSACTION_LIMIT_ID):
            try:
                await self._configure_gas_fee_for_transaction(transaction=transaction)
            except RuntimeError as simulation_ex:
                if CONSTANTS.ACCOUNT_SEQUENCE_MISMATCH_ERROR in str(simulation_ex):
                    await self.initialize_trading_account()
                raise

        transaction.with_memo("")
        transaction.with_timeout_height(await self.timeout_height())

        signed_transaction_data = self._sign_and_encode(transaction=transaction)

        async with self.throttler.execute_task(limit_id=CONSTANTS.SEND_TRANSACTION):
            result = await self.query_executor.send_tx_sync_mode(tx_byte=signed_transaction_data)

        if CONSTANTS.ACCOUNT_SEQUENCE_MISMATCH_ERROR in result.get("rawLog", ""):
            await self.initialize_trading_account()

        return result

    def _chain_stream_exception_handler(self, exception: RpcError):
        self.logger().warning(f"Error while listening to chain stream ({exception})")

    def _chain_stream_closed_handler(self):
        self.logger().debug("Reconnecting stream for chain stream")

    async def _listen_to_chain_updates(
            self,
            spot_markets: List[InjectiveSpotMarket],
            derivative_markets: List[InjectiveDerivativeMarket],
            subaccount_ids: List[str],
    ):
        composer = await self.composer()

        async def _chain_stream_event_handler(event: Dict[str, Any]):
            try:
                await self._process_chain_stream_update(
                    chain_stream_update=event, derivative_markets=derivative_markets,
                )
            except asyncio.CancelledError:
                raise
            except Exception as ex:
                self.logger().warning(f"Invalid chain stream event format ({ex})\n{event}")

        while True:
            # Running in a cycle to reconnect to the stream after connection errors
            await self._listen_chain_stream_updates(
                spot_markets=spot_markets,
                derivative_markets=derivative_markets,
                subaccount_ids=subaccount_ids,
                composer=composer,
                callback=_chain_stream_event_handler,
                on_end_callback=self._chain_stream_closed_handler,
                on_status_callback=self._chain_stream_exception_handler,
            )

    def _transaction_stream_exception_handler(self, exception: RpcError):
        self.logger().warning(f"Error while listening to transaction stream ({exception})")

    def _transaction_stream_closed_handler(self):
        self.logger().debug("Reconnecting stream for transaction stream")

    async def _listen_to_chain_transactions(self):
        while True:
            # Running in a cycle to reconnect to the stream after connection errors
            await self._listen_transactions_updates(
                callback=self._process_transaction_update,
                on_end_callback=self._transaction_stream_closed_handler,
                on_status_callback=self._transaction_stream_exception_handler,
            )

    async def _process_chain_stream_update(
        self, chain_stream_update: Dict[str, Any], derivative_markets: List[InjectiveDerivativeMarket],
    ):
        block_height = int(chain_stream_update["blockHeight"])
        block_timestamp = int(chain_stream_update["blockTime"]) * 1e-3
        tasks = []

        tasks.append(
            asyncio.create_task(
                self._process_subaccount_balance_update(
                    balance_events=chain_stream_update.get("subaccountDeposits", []),
                    block_height=block_height,
                    block_timestamp=block_timestamp,
                )
            )
        )
        tasks.append(
            asyncio.create_task(
                self._process_chain_spot_order_book_update(
                    order_book_updates=chain_stream_update.get("spotOrderbookUpdates", []),
                    block_height=block_height,
                    block_timestamp=block_timestamp,
                )
            )
        )
        tasks.append(
            asyncio.create_task(
                self._process_chain_spot_trade_update(
                    trade_updates=chain_stream_update.get("spotTrades", []),
                    block_height=block_height,
                    block_timestamp=block_timestamp,
                )
            )
        )
        tasks.append(
            asyncio.create_task(
                self._process_chain_derivative_order_book_update(
                    order_book_updates=chain_stream_update.get("derivativeOrderbookUpdates", []),
                    block_height=block_height,
                    block_timestamp=block_timestamp,
                )
            )
        )
        tasks.append(
            asyncio.create_task(
                self._process_chain_derivative_trade_update(
                    trade_updates=chain_stream_update.get("derivativeTrades", []),
                    block_height=block_height,
                    block_timestamp=block_timestamp,
                )
            )
        )
        tasks.append(
            asyncio.create_task(
                self._process_chain_order_update(
                    order_updates=chain_stream_update.get("spotOrders", []),
                    block_height = block_height,
                    block_timestamp = block_timestamp,
                )
            )
        )
        tasks.append(
            asyncio.create_task(
                self._process_chain_order_update(
                    order_updates=chain_stream_update.get("derivativeOrders", []),
                    block_height=block_height,
                    block_timestamp=block_timestamp,
                )
            )
        )
        tasks.append(
            asyncio.create_task(
                self._process_chain_position_updates(
                    position_updates=chain_stream_update.get("positions", []),
                    block_height=block_height,
                    block_timestamp=block_timestamp,
                )
            )
        )
        tasks.append(
            asyncio.create_task(
                self._process_oracle_price_updates(
                    oracle_price_updates=chain_stream_update.get("oraclePrices", []),
                    block_height=block_height,
                    block_timestamp=block_timestamp,
                    derivative_markets=derivative_markets,
                )
            )
        )

        await safe_gather(*tasks)

    async def _process_chain_spot_order_book_update(
            self,
            order_book_updates: List[Dict[str, Any]],
            block_height: int,
            block_timestamp: float
    ):
        for order_book_update in order_book_updates:
            try:
                market_id = order_book_update["orderbook"]["marketId"]
                market_info = await self.spot_market_info_for_id(market_id=market_id)
                await self._process_chain_order_book_update(
                    order_book_update=order_book_update,
                    block_height=block_height,
                    block_timestamp=block_timestamp,
                    market=market_info,
                )
            except asyncio.CancelledError:
                raise
            except Exception as ex:
                self.logger().warning(f"Error processing spot orderbook event ({ex})")
                self.logger().debug(f"Error processing the spot orderbook event {order_book_update}")

    async def _process_chain_derivative_order_book_update(
            self,
            order_book_updates: List[Dict[str, Any]],
            block_height: int,
            block_timestamp: float
    ):
        for order_book_update in order_book_updates:
            try:
                market_id = order_book_update["orderbook"]["marketId"]
                market_info = await self.derivative_market_info_for_id(market_id=market_id)
                await self._process_chain_order_book_update(
                    order_book_update=order_book_update,
                    block_height=block_height,
                    block_timestamp=block_timestamp,
                    market=market_info,
                )
            except asyncio.CancelledError:
                raise
            except Exception as ex:
                self.logger().warning(f"Error processing derivative orderbook event ({ex})")
                self.logger().debug(f"Error processing the derivative orderbook event {order_book_update}")

    async def _process_chain_order_book_update(
        self,
        order_book_update: Dict[str, Any],
        block_height: int,
        block_timestamp: float,
        market: Union[InjectiveSpotMarket, InjectiveDerivativeMarket],
    ):
        trading_pair = await self.trading_pair_for_market(market_id=market.market_id)
        buy_levels = sorted(
            order_book_update["orderbook"].get("buyLevels", []),
            key=lambda bid: int(bid["p"]),
            reverse=True
        )
        bids = [(market.price_from_special_chain_format(chain_price=Decimal(bid["p"])),
                 market.quantity_from_special_chain_format(chain_quantity=Decimal(bid["q"])))
                for bid in buy_levels]
        asks = [(market.price_from_special_chain_format(chain_price=Decimal(ask["p"])),
                 market.quantity_from_special_chain_format(chain_quantity=Decimal(ask["q"])))
                for ask in order_book_update["orderbook"].get("sellLevels", [])]

        order_book_message_content = {
            "trading_pair": trading_pair,
            "update_id": int(order_book_update["seq"]),
            "bids": bids,
            "asks": asks,
        }
        diff_message = OrderBookMessage(
            message_type=OrderBookMessageType.DIFF,
            content=order_book_message_content,
            timestamp=block_timestamp,
        )
        self.publisher.trigger_event(
            event_tag=OrderBookDataSourceEvent.DIFF_EVENT, message=diff_message
        )

    async def _process_chain_spot_trade_update(
        self,
        trade_updates: List[Dict[str, Any]],
        block_height: int,
        block_timestamp: float
    ):
        for trade_update in trade_updates:
            try:
                market_id = trade_update["marketId"]
                market_info = await self.spot_market_info_for_id(market_id=market_id)

                trading_pair = await self.trading_pair_for_market(market_id=market_id)
                timestamp = self._time()
                trade_type = TradeType.BUY if trade_update.get("isBuy", False) else TradeType.SELL
                amount = market_info.quantity_from_special_chain_format(
                    chain_quantity=Decimal(str(trade_update["quantity"]))
                )
                price = market_info.price_from_special_chain_format(chain_price=Decimal(str(trade_update["price"])))
                order_hash = to_0x_hex(base64.b64decode(trade_update["orderHash"]))
                client_order_id = trade_update.get("cid", "")
                trade_id = trade_update["tradeId"]
                message_content = {
                    "trade_id": trade_id,
                    "trading_pair": trading_pair,
                    "trade_type": float(trade_type.value),
                    "amount": amount,
                    "price": price,
                }
                trade_message = OrderBookMessage(
                    message_type=OrderBookMessageType.TRADE,
                    content=message_content,
                    timestamp=timestamp,
                )
                self.publisher.trigger_event(
                    event_tag=OrderBookDataSourceEvent.TRADE_EVENT, message=trade_message
                )

                fee_amount = market_info.quote_token.value_from_special_chain_format(chain_value=Decimal(trade_update["fee"]))
                fee = TradeFeeBase.new_spot_fee(
                    fee_schema=TradeFeeSchema(),
                    trade_type=trade_type,
                    percent_token=market_info.quote_token.symbol,
                    flat_fees=[TokenAmount(amount=fee_amount, token=market_info.quote_token.symbol)]
                )

                trade_update = TradeUpdate(
                    trade_id=trade_id,
                    client_order_id=client_order_id,
                    exchange_order_id=order_hash,
                    trading_pair=trading_pair,
                    fill_timestamp=timestamp,
                    fill_price=price,
                    fill_base_amount=amount,
                    fill_quote_amount=amount * price,
                    fee=fee,
                )
                self.publisher.trigger_event(event_tag=MarketEvent.TradeUpdate, message=trade_update)
            except asyncio.CancelledError:
                raise
            except Exception as ex:
                self.logger().warning(f"Error processing spot trade event ({ex})")
                self.logger().debug(f"Error processing the spot trade event {trade_update}")

    async def _process_chain_derivative_trade_update(
        self,
        trade_updates: List[Dict[str, Any]],
        block_height: int,
        block_timestamp: float
    ):
        for trade_update in trade_updates:
            try:
                market_id = trade_update["marketId"]
                market_info = await self.derivative_market_info_for_id(market_id=market_id)

                trading_pair = await self.trading_pair_for_market(market_id=market_id)
                trade_type = TradeType.BUY if trade_update.get("isBuy", False) else TradeType.SELL
                amount = market_info.quantity_from_special_chain_format(
                    chain_quantity=Decimal(str(trade_update["positionDelta"]["executionQuantity"]))
                )
                price = market_info.price_from_special_chain_format(
                    chain_price=Decimal(str(trade_update["positionDelta"]["executionPrice"])))
                order_hash = to_0x_hex(base64.b64decode(trade_update["orderHash"]))
                client_order_id = trade_update.get("cid", "")
                trade_id = trade_update["tradeId"]

                message_content = {
                    "trade_id": trade_id,
                    "trading_pair": trading_pair,
                    "trade_type": float(trade_type.value),
                    "amount": amount,
                    "price": price,
                }
                trade_message = OrderBookMessage(
                    message_type=OrderBookMessageType.TRADE,
                    content=message_content,
                    timestamp=block_timestamp,
                )
                self.publisher.trigger_event(
                    event_tag=OrderBookDataSourceEvent.TRADE_EVENT, message=trade_message
                )

                fee_amount = market_info.quote_token.value_from_special_chain_format(chain_value=Decimal(trade_update["fee"]))
                fee = TradeFeeBase.new_perpetual_fee(
                    fee_schema=TradeFeeSchema(),
                    position_action=PositionAction.OPEN,  # will be changed by the exchange class
                    percent_token=market_info.quote_token.symbol,
                    flat_fees=[TokenAmount(amount=fee_amount, token=market_info.quote_token.symbol)]
                )

                trade_update = TradeUpdate(
                    trade_id=trade_id,
                    client_order_id=client_order_id,
                    exchange_order_id=order_hash,
                    trading_pair=trading_pair,
                    fill_timestamp=block_timestamp,
                    fill_price=price,
                    fill_base_amount=amount,
                    fill_quote_amount=amount * price,
                    fee=fee,
                )
                self.publisher.trigger_event(event_tag=MarketEvent.TradeUpdate, message=trade_update)
            except asyncio.CancelledError:
                raise
            except Exception as ex:
                self.logger().warning(f"Error processing derivative trade event ({ex})")
                self.logger().debug(f"Error processing the derivative trade event {trade_update}")

    async def _process_chain_order_update(
            self,
            order_updates: List[Dict[str, Any]],
            block_height: int,
            block_timestamp: float,
    ):
        for order_update in order_updates:
            try:
                exchange_order_id = to_0x_hex(base64.b64decode(order_update["orderHash"]))
                client_order_id = order_update.get("cid", "")
                trading_pair = await self.trading_pair_for_market(market_id=order_update["order"]["marketId"])

                status_update = OrderUpdate(
                    trading_pair=trading_pair,
                    update_timestamp=block_timestamp,
                    new_state=CONSTANTS.STREAM_ORDER_STATE_MAP[order_update["status"]],
                    client_order_id=client_order_id,
                    exchange_order_id=exchange_order_id,
                )

                self.publisher.trigger_event(event_tag=MarketEvent.OrderUpdate, message=status_update)
            except asyncio.CancelledError:
                raise
            except Exception as ex:
                self.logger().warning(f"Error processing order event ({ex})")
                self.logger().debug(f"Error processing the order event {order_update}")

    async def _process_chain_position_updates(
            self,
            position_updates: List[Dict[str, Any]],
            block_height: int,
            block_timestamp: float,
    ):
        for event in position_updates:
            try:
                market_id = event["marketId"]
                market = await self.derivative_market_info_for_id(market_id=market_id)
                trading_pair = await self.trading_pair_for_market(market_id=market_id)

                position_side = PositionSide.LONG if event["isLong"] else PositionSide.SHORT
                amount_sign = Decimal(-1) if position_side == PositionSide.SHORT else Decimal(1)
                entry_price = (market.price_from_special_chain_format(chain_price=Decimal(event["entryPrice"])))
                amount = (market.quantity_from_special_chain_format(chain_quantity=Decimal(event["quantity"])))
                margin = (market.price_from_special_chain_format(chain_price=Decimal(event["margin"])))
                oracle_price = await self._oracle_price(market_id=market_id)
                leverage = (amount * entry_price) / margin
                unrealized_pnl = (oracle_price - entry_price) * amount * amount_sign

                parsed_event = PositionUpdateEvent(
                    timestamp=block_timestamp,
                    trading_pair=trading_pair,
                    position_side=position_side,
                    unrealized_pnl=unrealized_pnl,
                    entry_price=entry_price,
                    amount=amount * amount_sign,
                    leverage=leverage,
                )

                self.publisher.trigger_event(event_tag=AccountEvent.PositionUpdate, message=parsed_event)
            except asyncio.CancelledError:
                raise
            except Exception as ex:
                self.logger().warning(f"Error processing position event ({ex})")
                self.logger().debug(f"Error processing the position event {event}")

    async def _process_oracle_price_updates(
            self,
            oracle_price_updates: List[Dict[str, Any]],
            block_height: int,
            block_timestamp: float,
            derivative_markets: List[InjectiveDerivativeMarket],
    ):
        updated_symbols = {update["symbol"] for update in oracle_price_updates}
        for market in derivative_markets:
            try:
                if market.oracle_base() in updated_symbols or market.oracle_quote() in updated_symbols:
                    market_id = market.market_id
                    trading_pair = await self.trading_pair_for_market(market_id=market_id)
                    funding_info = await self.funding_info(market_id=market_id)
                    funding_info_update = FundingInfoUpdate(
                        trading_pair=trading_pair,
                        index_price=funding_info.index_price,
                        mark_price=funding_info.mark_price,
                        next_funding_utc_timestamp=funding_info.next_funding_utc_timestamp,
                        rate=funding_info.rate,
                    )
                    self.publisher.trigger_event(event_tag=MarketEvent.FundingInfo, message=funding_info_update)
            except asyncio.CancelledError:
                raise
            except Exception as ex:
                self.logger().warning(
                    f"Error processing oracle price update for market {market.trading_pair()} ({ex})"
                )

    async def _process_position_update(self, position_event: Dict[str, Any]):
        parsed_event = await self._parse_position_update_event(event=position_event)
        self.publisher.trigger_event(event_tag=AccountEvent.PositionUpdate, message=parsed_event)

    async def _process_subaccount_balance_update(
            self,
            balance_events: List[Dict[str, Any]],
            block_height: int,
            block_timestamp: float
    ):
        if len(balance_events) > 0 and self._uses_default_portfolio_subaccount():
            token_balances = await self.all_account_balances()

        for balance_event in balance_events:
            try:
                for deposit in balance_event["deposits"]:
                    updated_token = await self.token(denom=deposit["denom"])
                    if updated_token is not None:
                        if self._uses_default_portfolio_subaccount():
                            total_balance = token_balances[updated_token.unique_symbol]["total_balance"]
                            available_balance = token_balances[updated_token.unique_symbol]["available_balance"]
                        else:
                            updated_total = deposit["deposit"].get("totalBalance")
                            total_balance = (updated_token.value_from_special_chain_format(chain_value=Decimal(updated_total))
                                             if updated_total is not None
                                             else None)
                            updated_available = deposit["deposit"].get("availableBalance")
                            available_balance = (updated_token.value_from_special_chain_format(chain_value=Decimal(updated_available))
                                                 if updated_available is not None
                                                 else None)

                        balance_msg = BalanceUpdateEvent(
                            timestamp=self._time(),
                            asset_name=updated_token.unique_symbol,
                            total_balance=total_balance,
                            available_balance=available_balance,
                        )
                        self.publisher.trigger_event(event_tag=AccountEvent.BalanceEvent, message=balance_msg)
            except asyncio.CancelledError:
                raise
            except Exception as ex:
                self.logger().warning(f"Error processing subaccount balance event ({ex})")
                self.logger().debug(f"Error processing the subaccount balance event {balance_event}")

    async def _process_transaction_update(self, transaction_event: Dict[str, Any]):
        self.publisher.trigger_event(event_tag=InjectiveEvent.ChainTransactionEvent, message=transaction_event)

    async def _create_spot_order_definition(self, order: GatewayInFlightOrder):
        order_type = "BUY" if order.trade_type == TradeType.BUY else "SELL"
        if order.order_type == OrderType.LIMIT_MAKER:
            order_type = order_type + "_PO"
        composer = await self.composer()
        market_id = await self.market_id_for_spot_trading_pair(order.trading_pair)
        definition = composer.spot_order(
            market_id=market_id,
            subaccount_id=self.portfolio_account_subaccount_id,
            fee_recipient=self.portfolio_account_injective_address,
            price=order.price,
            quantity=order.amount,
            order_type=order_type,
            cid=order.client_order_id,
        )
        return definition

    async def _create_derivative_order_definition(self, order: GatewayPerpetualInFlightOrder):
        order_type = "BUY" if order.trade_type == TradeType.BUY else "SELL"
        if order.order_type == OrderType.LIMIT_MAKER:
            order_type = order_type + "_PO"
        composer = await self.composer()
        market_id = await self.market_id_for_derivative_trading_pair(order.trading_pair)
        definition = composer.derivative_order(
            market_id=market_id,
            subaccount_id=self.portfolio_account_subaccount_id,
            fee_recipient=self.portfolio_account_injective_address,
            price=order.price,
            quantity=order.amount,
            margin=composer.calculate_margin(
                quantity=order.amount,
                price=order.price,
                leverage=Decimal(str(order.leverage)),
                is_reduce_only=order.position == PositionAction.CLOSE,
            ),
            order_type=order_type,
            cid=order.client_order_id,
        )
        return definition

    def _create_trading_rules(
            self, markets: List[Union[InjectiveSpotMarket, InjectiveDerivativeMarket]]
    ) -> List[TradingRule]:
        trading_rules = []
        for market in markets:
            try:
                min_price_tick_size = market.min_price_tick_size()
                min_quantity_tick_size = market.min_quantity_tick_size()
                min_notional = market.min_notional()
                trading_rule = TradingRule(
                    trading_pair=market.trading_pair(),
                    min_order_size=min_quantity_tick_size,
                    min_price_increment=min_price_tick_size,
                    min_base_amount_increment=min_quantity_tick_size,
                    min_quote_amount_increment=min_price_tick_size,
                    min_notional_size=min_notional
                )
                trading_rules.append(trading_rule)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception(f"Error parsing the trading pair rule: {market.native_market}. Skipping...")

        return trading_rules

    async def _create_trading_fees(
            self, markets: List[Union[InjectiveSpotMarket, InjectiveDerivativeMarket]]
    ) -> Dict[str, TradeFeeSchema]:
        fees = {}
        for market in markets:
            trading_pair = await self.trading_pair_for_market(market_id=market.market_id)
            fees[trading_pair] = TradeFeeSchema(
                percent_fee_token=market.quote_token.unique_symbol,
                maker_percent_fee_decimal=market.maker_fee_rate(),
                taker_percent_fee_decimal=market.taker_fee_rate(),
            )

        return fees

    async def _get_markets_and_tokens(
        self
    ) -> Tuple[
        Dict[str, InjectiveToken],
        Mapping[str, str],
        Dict[str, InjectiveSpotMarket],
        Mapping[str, str],
        Dict[str, InjectiveDerivativeMarket],
        Mapping[str, str]
    ]:
        tokens_map = {}
        token_symbol_and_denom_map = bidict()
        spot_markets_map = {}
        derivative_markets_map = {}
        spot_market_id_to_trading_pair = bidict()
        derivative_market_id_to_trading_pair = bidict()

        async with self.throttler.execute_task(limit_id=CONSTANTS.SPOT_MARKETS_LIMIT_ID):
            async with self.throttler.execute_task(limit_id=CONSTANTS.DERIVATIVE_MARKETS_LIMIT_ID):
                spot_markets: Dict[str, SpotMarket] = await self.query_executor.spot_markets()
                derivative_markets: Dict[str, DerivativeMarket] = await self.query_executor.derivative_markets()
                tokens: Dict[str, Token] = await self.query_executor.tokens()

        for unique_symbol, injective_native_token in tokens.items():
            token = InjectiveToken(
                unique_symbol=unique_symbol,
                native_token=injective_native_token
            )
            tokens_map[token.denom] = token
            token_symbol_and_denom_map[unique_symbol] = token.denom

        for market in spot_markets.values():
            try:
                parsed_market = InjectiveSpotMarket(
                    market_id=market.id,
                    base_token=tokens_map[market.base_token.denom],
                    quote_token=tokens_map[market.quote_token.denom],
                    native_market=market
                )

                spot_market_id_to_trading_pair[parsed_market.market_id] = parsed_market.trading_pair()
                spot_markets_map[parsed_market.market_id] = parsed_market
            except KeyError:
                self.logger().debug(f"The spot market {market.id} will be excluded because it could not "
                                    f"be parsed ({market})")
                continue

        for market in derivative_markets.values():
            try:
                parsed_market = InjectiveDerivativeMarket(
                    market_id=market.id,
                    quote_token=tokens_map[market.quote_token.denom],
                    native_market=market,
                )

                if parsed_market.trading_pair() in derivative_market_id_to_trading_pair.inverse:
                    self.logger().debug(
                        f"The derivative market {market.id} will be excluded because there is other"
                        f" market with trading pair {parsed_market.trading_pair()} ({market})")
                    continue
                derivative_market_id_to_trading_pair[parsed_market.market_id] = parsed_market.trading_pair()
                derivative_markets_map[parsed_market.market_id] = parsed_market
            except KeyError:
                self.logger().debug(f"The derivative market {market.id} will be excluded because it could"
                                    f" not be parsed ({market})")
                continue

        return (
            tokens_map,
            token_symbol_and_denom_map,
            spot_markets_map,
            spot_market_id_to_trading_pair,
            derivative_markets_map,
            derivative_market_id_to_trading_pair
        )

    def _time(self):
        return time.time()

    async def _sleep(self, delay: float):
        """
        Method created to enable tests to prevent processes from sleeping
        """
        await asyncio.sleep(delay)



import logging
from inspect import getmembers, isabstract, isclass
from pathlib import Path
from shutil import copyfile, move

import pandas as pd
from sqlalchemy.exc import SQLAlchemyError

from hummingbot.client.config.config_helpers import ClientConfigAdapter
from hummingbot.model.db_migration.base_transformation import DatabaseTransformation
from hummingbot.model.sql_connection_manager import SQLConnectionManager, SQLConnectionType


class Migrator:
    @classmethod
    def _get_transformations(cls):
        import hummingbot.model.db_migration.transformations as transformations
        return [o for _, o in getmembers(transformations,
                                         predicate=lambda c: isclass(c) and
                                         issubclass(c, DatabaseTransformation) and
                                         not isabstract(c))]

    def __init__(self):
        self.transformations = [t(self) for t in self._get_transformations()]

    def migrate_db_to_version(self, client_config_map: ClientConfigAdapter, db_handle, from_version, to_version):
        original_db_path = db_handle.db_path
        original_db_name = Path(original_db_path).stem
        backup_db_path = original_db_path + '.backup_' + pd.Timestamp.utcnow().strftime("%Y%m%d-%H%M%S")
        new_db_path = original_db_path + '.new'
        copyfile(original_db_path, new_db_path)
        copyfile(original_db_path, backup_db_path)

        db_handle.engine.dispose()
        new_db_handle = SQLConnectionManager(
            client_config_map, SQLConnectionType.TRADE_FILLS, new_db_path, original_db_name, True
        )

        relevant_transformations = [t for t in self.transformations
                                    if t.does_apply_to_version(from_version, to_version)]
        if relevant_transformations:
            logging.getLogger().info(
                f"Will run DB migration from {from_version} to {to_version}")

        migration_successful = False
        try:
            for transformation in sorted(relevant_transformations):
                logging.getLogger().info(f"Applying {transformation.name} to DB...")
                new_db_handle = transformation.apply(new_db_handle)
                logging.getLogger().info(f"DONE with {transformation.name}")
            migration_successful = True
        except SQLAlchemyError:
            logging.getLogger().error("Unexpected error while checking and upgrading the local database.",
                                      exc_info=True)
        finally:
            try:
                new_db_handle.engine.dispose()
                if migration_successful:
                    move(new_db_path, original_db_path)
                db_handle.__init__(SQLConnectionType.TRADE_FILLS, original_db_path, original_db_name, True)
            except Exception as e:
                logging.getLogger().error(f"Fatal error migrating DB {original_db_path}")
                raise e
        return migration_successful



import asyncio
import time
from typing import TYPE_CHECKING, List, Optional

from hummingbot.connector.exchange.mexc import mexc_constants as CONSTANTS, mexc_web_utils as web_utils
from hummingbot.connector.exchange.mexc.mexc_auth import MexcAuth
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.mexc.mexc_exchange import MexcExchange


class MexcAPIUserStreamDataSource(UserStreamTrackerDataSource):
    LISTEN_KEY_KEEP_ALIVE_INTERVAL = 1800  # Recommended to Ping/Update listen key to keep connection alive
    HEARTBEAT_TIME_INTERVAL = 30.0

    _logger: Optional[HummingbotLogger] = None

    def __init__(self,
                 auth: MexcAuth,
                 trading_pairs: List[str],
                 connector: 'MexcExchange',
                 api_factory: WebAssistantsFactory,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN):
        super().__init__()
        self._auth: MexcAuth = auth
        self._current_listen_key = None
        self._domain = domain
        self._api_factory = api_factory

        self._listen_key_initialized_event: asyncio.Event = asyncio.Event()
        self._last_listen_key_ping_ts = 0

    async def _connected_websocket_assistant(self) -> WSAssistant:
        """
        Creates an instance of WSAssistant connected to the exchange
        """
        self._manage_listen_key_task = safe_ensure_future(self._manage_listen_key_task_loop())
        await self._listen_key_initialized_event.wait()

        ws: WSAssistant = await self._get_ws_assistant()
        url = f"{CONSTANTS.WSS_URL.format(self._domain)}?listenKey={self._current_listen_key}"
        await ws.connect(ws_url=url, ping_timeout=CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)
        return ws

    async def _subscribe_channels(self, websocket_assistant: WSAssistant):
        """
        Subscribes to order events and balance events.

        :param websocket_assistant: the websocket assistant used to connect to the exchange
        """
        try:

            orders_change_payload = {
                "method": "SUBSCRIPTION",
                "params": [CONSTANTS.USER_ORDERS_ENDPOINT_NAME],
                "id": 1
            }
            subscribe_order_change_request: WSJSONRequest = WSJSONRequest(payload=orders_change_payload)

            trades_payload = {
                "method": "SUBSCRIPTION",
                "params": [CONSTANTS.USER_TRADES_ENDPOINT_NAME],
                "id": 2
            }
            subscribe_trades_request: WSJSONRequest = WSJSONRequest(payload=trades_payload)

            balance_payload = {
                "method": "SUBSCRIPTION",
                "params": [CONSTANTS.USER_BALANCE_ENDPOINT_NAME],
                "id": 3
            }
            subscribe_balance_request: WSJSONRequest = WSJSONRequest(payload=balance_payload)

            await websocket_assistant.send(subscribe_order_change_request)
            await websocket_assistant.send(subscribe_trades_request)
            await websocket_assistant.send(subscribe_balance_request)

            self.logger().info("Subscribed to private order changes and balance updates channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to user streams...")
            raise

    async def _get_listen_key(self):
        rest_assistant = await self._api_factory.get_rest_assistant()
        try:
            data = await rest_assistant.execute_request(
                url=web_utils.public_rest_url(path_url=CONSTANTS.MEXC_USER_STREAM_PATH_URL, domain=self._domain),
                method=RESTMethod.POST,
                throttler_limit_id=CONSTANTS.MEXC_USER_STREAM_PATH_URL,
                is_auth_required=True
            )
        except asyncio.CancelledError:
            raise
        except Exception as exception:
            raise IOError(f"Error fetching user stream listen key. Error: {exception}")

        return data["listenKey"]

    async def _ping_listen_key(self) -> bool:
        rest_assistant = await self._api_factory.get_rest_assistant()
        try:
            data = await rest_assistant.execute_request(
                url=web_utils.public_rest_url(path_url=CONSTANTS.MEXC_USER_STREAM_PATH_URL, domain=self._domain),
                params={"listenKey": self._current_listen_key},
                method=RESTMethod.PUT,
                return_err=True,
                throttler_limit_id=CONSTANTS.MEXC_USER_STREAM_PATH_URL,
                is_auth_required=True
            )

            if "code" in data:
                self.logger().warning(f"Failed to refresh the listen key {self._current_listen_key}: {data}")
                return False

        except asyncio.CancelledError:
            raise
        except Exception as exception:
            self.logger().warning(f"Failed to refresh the listen key {self._current_listen_key}: {exception}")
            return False

        return True

    async def _manage_listen_key_task_loop(self):
        try:
            while True:
                now = int(time.time())
                if self._current_listen_key is None:
                    self._current_listen_key = await self._get_listen_key()
                    self.logger().info(f"Successfully obtained listen key {self._current_listen_key}")
                    self._listen_key_initialized_event.set()
                    self._last_listen_key_ping_ts = int(time.time())

                if now - self._last_listen_key_ping_ts >= self.LISTEN_KEY_KEEP_ALIVE_INTERVAL:
                    success: bool = await self._ping_listen_key()
                    if not success:
                        self.logger().error("Error occurred renewing listen key ...")
                        break
                    else:
                        self.logger().info(f"Refreshed listen key {self._current_listen_key}.")
                        self._last_listen_key_ping_ts = int(time.time())
                else:
                    await self._sleep(self.LISTEN_KEY_KEEP_ALIVE_INTERVAL)
        finally:
            self._current_listen_key = None
            self._listen_key_initialized_event.clear()

    async def _get_ws_assistant(self) -> WSAssistant:
        if self._ws_assistant is None:
            self._ws_assistant = await self._api_factory.get_ws_assistant()
        return self._ws_assistant

    async def _send_ping(self, websocket_assistant: WSAssistant):
        payload = {
            "method": "PING",
        }
        ping_request: WSJSONRequest = WSJSONRequest(payload=payload)
        await websocket_assistant.send(ping_request)

    async def _on_user_stream_interruption(self, websocket_assistant: Optional[WSAssistant]):
        await super()._on_user_stream_interruption(websocket_assistant=websocket_assistant)
        self._manage_listen_key_task and self._manage_listen_key_task.cancel()
        self._current_listen_key = None
        self._listen_key_initialized_event.clear()
        await self._sleep(5)

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant, queue: asyncio.Queue):
        while True:
            try:
                await asyncio.wait_for(
                    super()._process_websocket_messages(websocket_assistant=websocket_assistant, queue=queue),
                    timeout=CONSTANTS.WS_CONNECTION_TIME_INTERVAL
                )
            except asyncio.TimeoutError:
                ping_request = WSJSONRequest(payload={"method": "PING"})
                await websocket_assistant.send(ping_request)



# A single source of truth for constant variables related to the exchange

from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.in_flight_order import OrderState

EXCHANGE_NAME = "htx"
BROKER_ID = "AAc484720a"
DOMAIN = ""
MAX_CLIENT_ORDER_ID_LENGTH = 64


REST_URL = "https://api.huobi.pro"
WS_PUBLIC_URL = "wss://api.huobi.pro/ws"
WS_PRIVATE_URL = "wss://api.huobi.pro/ws/v2"

WS_HEARTBEAT_TIME_INTERVAL = 5  # seconds

# Websocket event types
TRADE_CHANNEL_SUFFIX = "trade.detail"
ORDERBOOK_CHANNEL_SUFFIX = "depth.step0"

TRADE_INFO_URL = "/v1/settings/common/market-symbols"
MOST_RECENT_TRADE_URL = "/market/tickers"
DEPTH_URL = "/market/depth"
LAST_TRADE_URL = "/market/trade"

SERVER_TIME_URL = "/v1/common/timestamp"
ACCOUNT_ID_URL = "/v1/account/accounts"
ACCOUNT_BALANCE_URL = "/v1/account/accounts/{}/balance"
OPEN_ORDERS_URL = "/v1/order/openOrders"
ORDER_DETAIL_URL = "/v1/order/orders/{}"
ORDER_MATCHES_URL = "/v1/order/orders/{}/matchresults"
PLACE_ORDER_URL = "/v1/order/orders/place"
CANCEL_ORDER_URL = "/v1/order/orders/{}/submitcancel"
BATCH_CANCEL_URL = "/v1/order/orders/batchcancel"

HTX_ACCOUNT_UPDATE_TOPIC = "accounts.update#2"
HTX_ORDER_UPDATE_TOPIC = "orders#{}"
HTX_TRADE_DETAILS_TOPIC = "trade.clearing#{}#0"

HTX_SUBSCRIBE_TOPICS = {HTX_ORDER_UPDATE_TOPIC, HTX_ACCOUNT_UPDATE_TOPIC, HTX_TRADE_DETAILS_TOPIC}

WS_CONNECTION_LIMIT_ID = "WSConnection"
WS_REQUEST_LIMIT_ID = "WSRequest"
CANCEL_URL_LIMIT_ID = "cancelRequest"
ACCOUNT_BALANCE_LIMIT_ID = "accountBalance"
ORDER_DETAIL_LIMIT_ID = "orderDetail"
ORDER_MATCHES_LIMIT_ID = "orderMatch"

RATE_LIMITS = [
    RateLimit(WS_CONNECTION_LIMIT_ID, limit=50, time_interval=1),
    RateLimit(WS_REQUEST_LIMIT_ID, limit=10, time_interval=1),
    RateLimit(limit_id=TRADE_INFO_URL, limit=10, time_interval=1),
    RateLimit(limit_id=MOST_RECENT_TRADE_URL, limit=10, time_interval=1),
    RateLimit(limit_id=DEPTH_URL, limit=10, time_interval=1),
    RateLimit(limit_id=LAST_TRADE_URL, limit=10, time_interval=1),
    RateLimit(limit_id=SERVER_TIME_URL, limit=10, time_interval=1),
    RateLimit(limit_id=ACCOUNT_ID_URL, limit=100, time_interval=2),
    RateLimit(limit_id=ACCOUNT_BALANCE_LIMIT_ID, limit=100, time_interval=2),
    RateLimit(limit_id=ORDER_DETAIL_LIMIT_ID, limit=50, time_interval=2),
    RateLimit(limit_id=ORDER_MATCHES_LIMIT_ID, limit=50, time_interval=2),
    RateLimit(limit_id=PLACE_ORDER_URL, limit=100, time_interval=2),
    RateLimit(limit_id=CANCEL_URL_LIMIT_ID, limit=100, time_interval=2),
    RateLimit(limit_id=BATCH_CANCEL_URL, limit=50, time_interval=2),

]

# Order States
ORDER_STATE = {
    "rejected": OrderState.FAILED,
    "canceled": OrderState.CANCELED,
    "submitted": OrderState.OPEN,
    "partial-filled": OrderState.PARTIALLY_FILLED,
    "filled": OrderState.FILLED,
    "partial-canceled": OrderState.CANCELED,
    "created": OrderState.PENDING_CREATE,
    "canceling": OrderState.PENDING_CANCEL
}



import asyncio
import copy
import logging
import math
import typing
from decimal import Decimal
from enum import Enum
from typing import Any, Dict, NamedTuple, Optional, Tuple

from async_timeout import timeout

from hummingbot.core.data_type.common import OrderType, PositionAction, TradeType
from hummingbot.core.data_type.limit_order import LimitOrder
from hummingbot.core.data_type.trade_fee import TradeFeeBase
from hummingbot.logger import HummingbotLogger

if typing.TYPE_CHECKING:  # avoid circular import problems
    from hummingbot.connector.exchange_base import ExchangeBase

s_decimal_0 = Decimal("0")

GET_EX_ORDER_ID_TIMEOUT = 10  # seconds


class OrderState(Enum):
    PENDING_CREATE = 0
    OPEN = 1
    PENDING_CANCEL = 2
    CANCELED = 3
    PARTIALLY_FILLED = 4
    FILLED = 5
    FAILED = 6
    PENDING_APPROVAL = 7
    APPROVED = 8
    CREATED = 9
    COMPLETED = 10


class OrderUpdate(NamedTuple):
    trading_pair: str
    update_timestamp: float  # seconds
    new_state: OrderState
    client_order_id: Optional[str] = None
    exchange_order_id: Optional[str] = None
    misc_updates: Optional[Dict[str, Any]] = None


class TradeUpdate(NamedTuple):
    trade_id: str
    client_order_id: str
    exchange_order_id: str
    trading_pair: str
    fill_timestamp: float  # seconds
    fill_price: Decimal
    fill_base_amount: Decimal
    fill_quote_amount: Decimal
    fee: TradeFeeBase
    is_taker: bool = True  # CEXs deliver trade events from the taker's perspective

    @property
    def fee_asset(self):
        return self.fee.fee_asset

    @classmethod
    def from_json(cls, data: Dict[str, Any]):
        instance = TradeUpdate(
            trade_id=data["trade_id"],
            client_order_id=data["client_order_id"],
            exchange_order_id=data["exchange_order_id"],
            trading_pair=data["trading_pair"],
            fill_timestamp=data["fill_timestamp"],
            fill_price=Decimal(data["fill_price"]),
            fill_base_amount=Decimal(data["fill_base_amount"]),
            fill_quote_amount=Decimal(data["fill_quote_amount"]),
            fee=TradeFeeBase.from_json(data["fee"]),
        )

        return instance

    def to_json(self) -> Dict[str, Any]:
        json_dict = self._asdict()
        json_dict.update({
            "fill_price": str(self.fill_price),
            "fill_base_amount": str(self.fill_base_amount),
            "fill_quote_amount": str(self.fill_quote_amount),
            "fee": self.fee.to_json(),
        })
        return json_dict


class InFlightOrder:
    _logger: Optional[HummingbotLogger] = None

    def __init__(
            self,
            client_order_id: str,
            trading_pair: str,
            order_type: OrderType,
            trade_type: TradeType,
            amount: Decimal,
            creation_timestamp: float,
            price: Optional[Decimal] = None,
            exchange_order_id: Optional[str] = None,
            initial_state: OrderState = OrderState.PENDING_CREATE,
            leverage: int = 1,
            position: PositionAction = PositionAction.NIL,
    ) -> None:
        self.client_order_id = client_order_id
        self.creation_timestamp = creation_timestamp
        self.trading_pair = trading_pair
        self.order_type = order_type
        self.trade_type = trade_type
        self.price = price
        self.amount = amount
        self.exchange_order_id = exchange_order_id
        self.current_state = initial_state
        self.leverage = leverage
        self.position = position

        self.executed_amount_base = s_decimal_0
        self.executed_amount_quote = s_decimal_0

        self.last_update_timestamp: float = creation_timestamp

        self.order_fills: Dict[str, TradeUpdate] = {}  # Dict[trade_id, TradeUpdate]

        self.exchange_order_id_update_event = asyncio.Event()
        if self.exchange_order_id:
            self.exchange_order_id_update_event.set()
        self.completely_filled_event = asyncio.Event()
        self.processed_by_exchange_event = asyncio.Event()
        self.check_processed_by_exchange_condition()

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    @property
    def attributes(self) -> Tuple[Any]:
        return copy.deepcopy(
            (
                self.client_order_id,
                self.trading_pair,
                self.order_type,
                self.trade_type,
                self.price,
                self.amount,
                self.exchange_order_id,
                self.current_state,
                self.leverage,
                self.position,
                self.executed_amount_base,
                self.executed_amount_quote,
                self.creation_timestamp,
                self.last_update_timestamp,
            )
        )

    def __eq__(self, other: object) -> bool:
        return type(self) is type(other) and self.attributes == other.attributes

    @property
    def base_asset(self):
        return self.trading_pair.split("-")[0]

    @property
    def quote_asset(self):
        return self.trading_pair.split("-")[1]

    @property
    def is_pending_create(self) -> bool:
        return self.current_state == OrderState.PENDING_CREATE

    @property
    def is_pending_cancel_confirmation(self) -> bool:
        return self.current_state == OrderState.PENDING_CANCEL

    @property
    def is_open(self) -> bool:
        return self.current_state in {
            OrderState.PENDING_CREATE,
            OrderState.OPEN,
            OrderState.PARTIALLY_FILLED,
            OrderState.PENDING_CANCEL}

    @property
    def is_done(self) -> bool:
        return (
            self.current_state in {OrderState.CANCELED, OrderState.FILLED, OrderState.FAILED}
            or math.isclose(self.executed_amount_base, self.amount)
            or self.executed_amount_base >= self.amount
        )

    @property
    def is_filled(self) -> bool:
        return (
            self.current_state == OrderState.FILLED
            or (self.amount != s_decimal_0
                and (math.isclose(self.executed_amount_base, self.amount)
                     or self.executed_amount_base >= self.amount)
                )
        )

    @property
    def is_failure(self) -> bool:
        return self.current_state == OrderState.FAILED

    @property
    def is_cancelled(self) -> bool:
        return self.current_state == OrderState.CANCELED

    @property
    def average_executed_price(self) -> Optional[Decimal]:
        executed_value: Decimal = s_decimal_0
        total_base_amount: Decimal = s_decimal_0
        for order_fill in self.order_fills.values():
            executed_value += order_fill.fill_price * order_fill.fill_base_amount
            total_base_amount += order_fill.fill_base_amount
        if executed_value == s_decimal_0 or total_base_amount == s_decimal_0:
            return None
        return executed_value / total_base_amount

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> "InFlightOrder":
        """
        Initialize an InFlightOrder using a JSON object
        :param data: JSON data
        :return: Formatted InFlightOrder
        """
        order = InFlightOrder(
            client_order_id=data["client_order_id"],
            trading_pair=data["trading_pair"],
            order_type=getattr(OrderType, data["order_type"]),
            trade_type=getattr(TradeType, data["trade_type"]),
            amount=Decimal(data["amount"]),
            price=Decimal(data["price"]),
            exchange_order_id=data["exchange_order_id"],
            initial_state=OrderState(int(data["last_state"])),
            leverage=int(data["leverage"]),
            position=PositionAction(data["position"]),
            creation_timestamp=data.get("creation_timestamp", -1)
        )
        order.executed_amount_base = Decimal(data["executed_amount_base"])
        order.executed_amount_quote = Decimal(data["executed_amount_quote"])
        order.order_fills.update({key: TradeUpdate.from_json(value)
                                  for key, value
                                  in data.get("order_fills", {}).items()})
        order.last_update_timestamp = data.get("last_update_timestamp", order.creation_timestamp)

        order.check_filled_condition()
        order.check_processed_by_exchange_condition()

        return order

    def to_json(self) -> Dict[str, Any]:
        """
        Returns this InFlightOrder as a JSON object.
        :return: JSON object
        """
        return {
            "client_order_id": self.client_order_id,
            "exchange_order_id": self.exchange_order_id,
            "trading_pair": self.trading_pair,
            "order_type": self.order_type.name,
            "trade_type": self.trade_type.name,
            "price": str(self.price),
            "amount": str(self.amount),
            "executed_amount_base": str(self.executed_amount_base),
            "executed_amount_quote": str(self.executed_amount_quote),
            "last_state": str(self.current_state.value),
            "leverage": str(self.leverage),
            "position": self.position.value,
            "creation_timestamp": self.creation_timestamp,
            "last_update_timestamp": self.last_update_timestamp,
            "order_fills": {key: fill.to_json() for key, fill in self.order_fills.items()},
            "cumulative_fee_paid_base": float(self.cumulative_fee_paid(self.base_asset)),
            "cumulative_fee_paid_quote": float(self.cumulative_fee_paid(self.quote_asset)),
        }

    def to_limit_order(self) -> LimitOrder:
        """
        Returns this InFlightOrder as a LimitOrder object.
        :return: LimitOrder object.
        """
        return LimitOrder(
            client_order_id=self.client_order_id,
            trading_pair=self.trading_pair,
            is_buy=self.trade_type is TradeType.BUY,
            base_currency=self.base_asset,
            quote_currency=self.quote_asset,
            price=self.price,
            quantity=self.amount,
            filled_quantity=self.executed_amount_base,
            creation_timestamp=int(self.creation_timestamp * 1e6)
        )

    def update_exchange_order_id(self, exchange_order_id: str):
        self.exchange_order_id = exchange_order_id
        self.exchange_order_id_update_event.set()

    async def get_exchange_order_id(self):
        if self.exchange_order_id is None:
            async with timeout(GET_EX_ORDER_ID_TIMEOUT):
                await self.exchange_order_id_update_event.wait()
        return self.exchange_order_id

    def cumulative_fee_paid(self, token: str, exchange: Optional['ExchangeBase'] = None) -> Decimal:
        """
        Returns the total amount of fee paid for each trade update, expressed in the specified token
        :param token: The token all partial fills' fees should be transformed to before summing them
        :param exchange: The exchange being used. If specified the logic will try to use the order book to get the rate
        :return: the cumulative fee paid for all partial fills in the specified token
        """
        total_fee_in_token = Decimal("0")
        try:
            for trade_update in self.order_fills.values():
                total_fee_in_token += trade_update.fee.fee_amount_in_token(
                    trading_pair=self.trading_pair,
                    price=trade_update.fill_price,
                    order_amount=trade_update.fill_base_amount,
                    token=token,
                    exchange=exchange
                )
        except Exception as e:
            self.logger().error(f"Error calculating fee paid in {token}: {e}")
        return total_fee_in_token

    def update_with_order_update(self, order_update: OrderUpdate) -> bool:
        """
        Updates the in flight order with an order update (from REST API or WS API)
        return: True if the order gets updated otherwise False
        """
        if (order_update.client_order_id != self.client_order_id
                and order_update.exchange_order_id != self.exchange_order_id):
            return False

        prev_data = (self.exchange_order_id, self.current_state)

        if self.exchange_order_id is None and order_update.exchange_order_id is not None:
            self.update_exchange_order_id(order_update.exchange_order_id)

        self.current_state = order_update.new_state
        self.check_processed_by_exchange_condition()

        updated: bool = prev_data != (self.exchange_order_id, self.current_state)

        if updated:
            self.last_update_timestamp = order_update.update_timestamp

        return updated

    def update_with_trade_update(self, trade_update: TradeUpdate) -> bool:
        """
        Updates the in flight order with a trade update (from REST API or WS API)
        :return: True if the order gets updated otherwise False
        """
        trade_id: str = trade_update.trade_id

        if (trade_id in self.order_fills
                or (self.client_order_id != trade_update.client_order_id
                    and self.exchange_order_id != trade_update.exchange_order_id)):
            return False

        self.order_fills[trade_id] = trade_update

        self.executed_amount_base += trade_update.fill_base_amount
        self.executed_amount_quote += trade_update.fill_quote_amount

        self.last_update_timestamp = trade_update.fill_timestamp
        self.check_filled_condition()

        return True

    def check_filled_condition(self):
        if (abs(self.amount) - self.executed_amount_base).quantize(Decimal('1e-8')) <= 0:
            self.completely_filled_event.set()

    async def wait_until_completely_filled(self):
        await self.completely_filled_event.wait()

    def check_processed_by_exchange_condition(self):
        if self.current_state.value > OrderState.PENDING_CREATE.value:
            self.processed_by_exchange_event.set()

    async def wait_until_processed_by_exchange(self):
        await self.processed_by_exchange_event.wait()

    def build_order_created_message(self) -> str:
        return (
            f"Created {self.order_type.name.upper()} {self.trade_type.name.upper()} order "
            f"{self.client_order_id} for {self.amount} {self.trading_pair} "
            f"at {self.price}."
        )


class PerpetualDerivativeInFlightOrder(InFlightOrder):
    def build_order_created_message(self) -> str:
        return (
            f"Created {self.order_type.name.upper()} {self.trade_type.name.upper()} order "
            f"{self.client_order_id} for {self.amount} to {self.position.name.upper()} a {self.trading_pair} position "
            f"at {self.price}."
        )



import time
from typing import Optional

import hummingbot.connector.exchange.kraken.kraken_constants as CONSTANTS
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


def private_rest_url(*args, **kwargs) -> str:
    return rest_url(*args, **kwargs)


def public_rest_url(*args, **kwargs) -> str:
    return rest_url(*args, **kwargs)


def rest_url(path_url: str, domain: str = "kraken"):
    base_url = CONSTANTS.BASE_URL
    return base_url + path_url


def build_api_factory(
        throttler: Optional[AsyncThrottler] = None,
        auth: Optional[AuthBase] = None, ) -> WebAssistantsFactory:
    throttler = throttler
    api_factory = WebAssistantsFactory(
        throttler=throttler,
        auth=auth
    )
    return api_factory


def is_exchange_information_valid(trading_pair_details) -> bool:
    """
    Verifies if a trading pair is enabled to operate with based on its exchange information

    :param exchange_info: the exchange information for a trading pair

    :return: True if the trading pair is enabled, False otherwise
    Want to filter out dark pool trading pairs from the list of trading pairs
    For more info, please check
    https://support.kraken.com/hc/en-us/articles/360001391906-Introducing-the-Kraken-Dark-Pool
    """
    if trading_pair_details.get('altname'):
        return not trading_pair_details.get('altname').endswith('.d')
    return True


async def get_current_server_time(
        throttler,
        domain
) -> float:
    return time.time()



import asyncio
import logging
from typing import Optional

import hummingbot.connector.exchange.tegro.tegro_web_utils as web_utils
from hummingbot.connector.exchange.tegro import tegro_constants as CONSTANTS
from hummingbot.connector.exchange.tegro.tegro_auth import TegroAuth
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

MESSAGE_TIMEOUT = 20.0
PING_TIMEOUT = 5.0


class TegroUserStreamDataSource(UserStreamTrackerDataSource):

    _bpusds_logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._bpusds_logger is None:
            cls._bpusds_logger = logging.getLogger(__name__)
        return cls._bpusds_logger

    def __init__(
        self,
        auth: TegroAuth,
        domain: str = CONSTANTS.DOMAIN,
        throttler: Optional[AsyncThrottler] = None,
        api_factory: Optional[WebAssistantsFactory] = None,
    ):
        super().__init__()
        self._domain = domain
        self._throttler = throttler
        self._api_factory: WebAssistantsFactory = api_factory or web_utils.build_api_factory(
            auth=auth
        )
        self._auth: TegroAuth = auth
        self._ws_assistant: Optional[WSAssistant] = None

    @property
    def last_recv_time(self) -> float:
        if self._ws_assistant:
            return self._ws_assistant.last_recv_time
        return 0

    async def _get_ws_assistant(self) -> WSAssistant:
        if self._ws_assistant is None:
            self._ws_assistant = await self._api_factory.get_ws_assistant()
        return self._ws_assistant

    async def _send_ping(self, websocket_assistant: WSAssistant):
        API_KEY = self._auth._api_key
        payload = {"action": "subscribe", "channelId": API_KEY}
        ping_request: WSJSONRequest = WSJSONRequest(payload=payload)
        await websocket_assistant.send(ping_request)

    async def listen_for_user_stream(self, output: asyncio.Queue):
        ws = None
        while True:
            try:
                # # establish initial connection to websocket
                ws: WSAssistant = await self._get_ws_assistant()
                await ws.connect(ws_url=web_utils.wss_url(CONSTANTS.PUBLIC_WS_ENDPOINT, self._domain), ping_timeout=PING_TIMEOUT)

                # # send auth request
                API_KEY = self._auth._api_key
                subscribe_payload = {"action": "subscribe", "channelId": API_KEY}

                subscribe_request: WSJSONRequest = WSJSONRequest(
                    payload=subscribe_payload,
                    is_auth_required=False
                )
                await ws.send(subscribe_request)
                await self._send_ping(ws)
                async for msg in ws.iter_messages():
                    if msg.data is not None and len(msg.data) > 0:
                        output.put_nowait(msg.data)
            except asyncio.CancelledError:
                raise
            except Exception as e:
                self.logger().error(
                    f"Unexpected error while listening to user stream. Retrying after 5 seconds... "
                    f"Error: {e}",
                    exc_info=True,
                )
            finally:
                # Make sure no background task is leaked.
                ws and await ws.disconnect()
                await self._sleep(5)



import hashlib
import hmac
import json
from typing import Any, Dict
from urllib.parse import urlparse

import six

from hummingbot.connector.exchange.gate_io import gate_io_constants as CONSTANTS
from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTRequest, WSRequest


class GateIoAuth(AuthBase):
    """
    Auth Gate.io API
    https://www.gate.io/docs/apiv4/en/#authentication
    """
    def __init__(self, api_key: str, secret_key: str, time_provider: TimeSynchronizer):
        self.api_key = api_key
        self.secret_key = secret_key
        self.time_provider = time_provider

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        headers = {}
        if request.headers is not None:
            headers.update(request.headers)
        headers.update(self._get_auth_headers(request))
        request.headers = headers
        return request

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        request.payload["auth"] = self._get_auth_headers_ws(payload=request.payload)
        return request

    def _get_auth_headers_ws(self, payload: Dict[str, Any] = None) -> Dict[str, Any]:
        """
        Generates authn for Gate.io websockets

        :return: a dictionary with headers
        """
        sig = self._sign_payload_ws(payload['channel'], payload['event'], payload['time'])
        headers = {
            "method": "api_key",
            "KEY": f"{self.api_key}",
            "SIGN": f"{sig}",
        }
        return headers

    def _get_auth_headers(self, request: RESTRequest) -> Dict[str, Any]:
        """
        Generates authentication headers for Gate.io REST API

        :return: a dictionary with headers
        """
        sign, ts = self._sign_payload(request)
        headers = {
            "KEY": f"{self.api_key}",
            "SIGN": f"{sign}",
            "Timestamp": f"{ts}",
            "Content-Type": "application/json",
            "X-Gate-Channel-Id": CONSTANTS.HBOT_BROKER_ID,
        }
        return headers

    def _sign_payload_ws(self, channel, event, time) -> str:
        return self._sign(f"channel={channel}&event={event}&time={time}")

    def _sign_payload(self, r: RESTRequest) -> tuple[str, int]:
        query_string = ""
        body = r.data

        ts = self.time_provider.time()
        m = hashlib.sha512()
        path = urlparse(r.url).path

        if body is not None:
            if not isinstance(r.data, six.string_types):
                body = json.dumps(r.data)
            m.update(body.encode('utf-8'))
        body_hash = m.hexdigest()

        if r.params:
            qs = []
            for k, v in r.params.items():
                qs.append(f"{k}={v}")
            query_string = "&".join(qs)

        s = f'{r.method}\n{path}\n{query_string}\n{body_hash}\n{ts}'
        return self._sign(s), ts

    def _sign(self, payload) -> str:
        return hmac.new(
            self.secret_key.encode('utf-8'),
            payload.encode('utf-8'),
            hashlib.sha512).hexdigest()



import json
from collections import OrderedDict
from typing import Any, Dict

from eth_account import Account, messages

from hummingbot.connector.utils import to_0x_hex
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, RESTRequest, WSRequest


class TegroAuth(AuthBase):
    """
    Auth class required by Tegro API
    """

    def __init__(self, api_key: str, api_secret: str):
        self._api_key: str = api_key
        self._api_secret: str = api_secret

    def sign_inner(self, data):
        """
        Sign the provided data using the API secret key.
        """
        wallet = Account.from_key(self._api_secret)
        return to_0x_hex(wallet.sign_message(data).signature)

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        """
        Adds the server time and the signature to the request, required for authenticated interactions. It also adds
        the required parameter in the request header.
        :param request: the request to be configured for authenticated interaction
        """
        if request.method == RESTMethod.POST and request.data is not None:
            request.data = self.add_auth_to_params(params=json.loads(request.data) if request.data is not None else {})
        else:
            request.params = self.add_auth_to_params(params=request.params)
        # Generates auth headers

        headers = {}
        if request.headers is not None:
            headers.update(request.headers)
        headers.update(self.header_for_authentication())
        request.headers = headers

        return request

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        return request  # pass-through

    def add_auth_to_params(self,
                           params: Dict[str, Any]):
        request_params = OrderedDict(params or {})

        addr = self._api_key
        address = addr.lower()
        structured_data = messages.encode_defunct(text=address)
        signature = self.sign_inner(structured_data)
        request_params["signature"] = signature
        return request_params

    def header_for_authentication(self) -> Dict[str, Any]:
        return {
            "Content-Type": 'application/json',
        }

    def get_auth_headers(self):
        headers = self.header_for_authentication()
        headers.update(self._generate_auth_dict())
        return headers



import time
from decimal import Decimal
from typing import Any, Dict, Optional, Tuple

import hummingbot.connector.exchange.hyperliquid.hyperliquid_constants as CONSTANTS
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTRequest
from hummingbot.core.web_assistant.rest_pre_processors import RESTPreProcessorBase
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


class HyperliquidPerpetualRESTPreProcessor(RESTPreProcessorBase):

    async def pre_process(self, request: RESTRequest) -> RESTRequest:
        if request.headers is None:
            request.headers = {}
        request.headers["Content-Type"] = (
            "application/json"
        )
        return request


def private_rest_url(*args, **kwargs) -> str:
    return rest_url(*args, **kwargs)


def public_rest_url(*args, **kwargs) -> str:
    return rest_url(*args, **kwargs)


def rest_url(path_url: str, domain: str = "hyperliquid"):
    base_url = CONSTANTS.BASE_URL if domain == "hyperliquid" else CONSTANTS.TESTNET_BASE_URL
    return base_url + path_url


def wss_url(domain: str = "hyperliquid"):
    base_ws_url = CONSTANTS.WS_URL if domain == "hyperliquid" else CONSTANTS.TESTNET_WS_URL
    return base_ws_url


def build_api_factory(
        throttler: Optional[AsyncThrottler] = None,
        auth: Optional[AuthBase] = None) -> WebAssistantsFactory:
    throttler = throttler or create_throttler()
    api_factory = WebAssistantsFactory(
        throttler=throttler,
        rest_pre_processors=[HyperliquidPerpetualRESTPreProcessor()],
        auth=auth)
    return api_factory


def build_api_factory_without_time_synchronizer_pre_processor(throttler: AsyncThrottler) -> WebAssistantsFactory:
    api_factory = WebAssistantsFactory(
        throttler=throttler,
        rest_pre_processors=[HyperliquidPerpetualRESTPreProcessor()])
    return api_factory


def create_throttler() -> AsyncThrottler:
    return AsyncThrottler(CONSTANTS.RATE_LIMITS)


async def get_current_server_time(
        throttler,
        domain
) -> float:
    return time.time()


def is_exchange_information_valid(rule: Dict[str, Any]) -> bool:
    """
    Verifies if a trading pair is enabled to operate with based on its exchange information

    :param exchange_info: the exchange information for a trading pair

    :return: True if the trading pair is enabled, False otherwise
    """
    return True


def order_type_to_tuple(order_type) -> Tuple[int, float]:
    if "limit" in order_type:
        tif = order_type["limit"]["tif"]
        if tif == "Gtc":
            return 2, 0
        elif tif == "Alo":
            return 1, 0
        elif tif == "Ioc":
            return 3, 0
    elif "trigger" in order_type:
        trigger = order_type["trigger"]
        trigger_px = trigger["triggerPx"]
        if trigger["isMarket"] and trigger["tpsl"] == "tp":
            return 4, trigger_px
        elif not trigger["isMarket"] and trigger["tpsl"] == "tp":
            return 5, trigger_px
        elif trigger["isMarket"] and trigger["tpsl"] == "sl":
            return 6, trigger_px
        elif not trigger["isMarket"] and trigger["tpsl"] == "sl":
            return 7, trigger_px
    raise ValueError("Invalid order type", order_type)


def float_to_int_for_hashing(x: float) -> int:
    return float_to_int(x, 8)


def float_to_int(x: float, power: int) -> int:
    with_decimals = x * 10 ** power
    if abs(round(with_decimals) - with_decimals) >= 1e-3:
        raise ValueError("float_to_int causes rounding", x)
    return round(with_decimals)


def str_to_bytes16(x: str) -> bytearray:
    assert x.startswith("0x")
    return bytearray.fromhex(x[2:])


def order_grouping_to_number(grouping) -> int:
    if grouping == "na":
        return 0
    elif grouping == "normalTpsl":
        return 1
    elif grouping == "positionTpsl":
        return 2


def order_spec_to_order_wire(order_spec):
    return {
        "a": order_spec["asset"],
        "b": order_spec["isBuy"],
        "p": float_to_wire(order_spec["limitPx"]),
        "s": float_to_wire(order_spec["sz"]),
        "r": order_spec["reduceOnly"],
        "t": order_type_to_wire(order_spec["orderType"]),
        "c": order_spec["cloid"],
    }


def float_to_wire(x: float) -> str:
    rounded = "{:.8f}".format(x)
    if abs(float(rounded) - x) >= 1e-12:
        raise ValueError("float_to_wire causes rounding", x)
    if rounded == "-0":
        rounded = "0"
    normalized = Decimal(rounded).normalize()
    return f"{normalized:f}"


def order_type_to_wire(order_type):
    if "limit" in order_type:
        return {"limit": order_type["limit"]}
    elif "trigger" in order_type:
        return {
            "trigger": {
                "triggerPx": float_to_wire(order_type["trigger"]["triggerPx"]),
                "tpsl": order_type["trigger"]["tpsl"],
                "isMarket": order_type["trigger"]["isMarket"],
            }
        }
    raise ValueError("Invalid order type", order_type)
